# Copyright (c) 2014 Alex Meade.  All rights reserved.
# Copyright (c) 2014 Clinton Knight.  All rights reserved.
# Copyright (c) 2015 Tom Barron.  All rights reserved.
# Copyright (c) 2016 Mike Rooney. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from oslo_log import log as logging
from oslo_utils import excutils

import six

from cinder import exception
from cinder.i18n import _
from cinder import utils
from cinder.volume.drivers.netapp.dataontap.client import api as netapp_api
from cinder.volume.drivers.netapp import utils as na_utils


LOG = logging.getLogger(__name__)

DELETED_PREFIX = 'deleted_cinder_'
MAX_SIZE_FOR_A_LUN = '17555678822400'


@six.add_metaclass(utils.TraceWrapperMetaclass)
class Client(object):

    def __init__(self, **kwargs):
        host = kwargs['hostname']
        username = kwargs['username']
        password = kwargs['password']
        api_trace_pattern = kwargs['api_trace_pattern']
        self.connection = netapp_api.NaServer(
            host=host,
            transport_type=kwargs['transport_type'],
            port=kwargs['port'],
            username=username,
            password=password,
            api_trace_pattern=api_trace_pattern)

        self.ssh_client = self._init_ssh_client(host, username, password)

    def _init_ssh_client(self, host, username, password):
        return netapp_api.SSHUtil(
            host=host,
            username=username,
            password=password)

    def _init_features(self):
        """Set up the repository of available Data ONTAP features."""
        self.features = na_utils.Features()

    def get_ontap_version(self, cached=True):
        """Gets the ONTAP version."""

        if cached:
            return self.connection.get_ontap_version()

        ontap_version = netapp_api.NaElement("system-get-version")
        result = self.connection.invoke_successfully(ontap_version, True)

        version_tuple = result.get_child_by_name(
            'version-tuple') or netapp_api.NaElement('none')
        system_version_tuple = version_tuple.get_child_by_name(
            'system-version-tuple') or netapp_api.NaElement('none')

        generation = system_version_tuple.get_child_content("generation")
        major = system_version_tuple.get_child_content("major")

        return '%(generation)s.%(major)s' % {
            'generation': generation,
            'major': major}

    def get_ontapi_version(self, cached=True):
        """Gets the supported ontapi version."""

        if cached:
            return self.connection.get_api_version()

        ontapi_version = netapp_api.NaElement('system-get-ontapi-version')
        res = self.connection.invoke_successfully(ontapi_version, False)
        major = res.get_child_content('major-version')
        minor = res.get_child_content('minor-version')
        return major, minor

    def _strip_xml_namespace(self, string):
        if string.startswith('{') and '}' in string:
            return string.split('}', 1)[1]
        return string

    def check_is_naelement(self, elem):
        """Checks if object is instance of NaElement."""
        if not isinstance(elem, netapp_api.NaElement):
            raise ValueError('Expects NaElement')

    def create_lun(self, volume_name, lun_name, size, metadata,
                   qos_policy_group_name=None):
        """Issues API request for creating LUN on volume."""

        path = '/vol/%s/%s' % (volume_name, lun_name)
        space_reservation = metadata['SpaceReserved']
        initial_size = size
        ontap_version = self.get_ontap_version()

        # On older ONTAP versions the extend size is limited to its
        # geometry on max_resize_size. In order to remove this
        # limitation we create the LUN with its maximum possible size
        # and then shrink to the requested size.
        if ontap_version < '9.5':
            initial_size = MAX_SIZE_FOR_A_LUN
            # In order to create a LUN with its maximum size (16TB),
            # the space_reservation needs to be disabled
            space_reservation = 'false'

        params = {'path': path, 'size': str(initial_size),
                  'ostype': metadata['OsType'],
                  'space-reservation-enabled': space_reservation}
        version = self.get_ontapi_version()
        if version >= (1, 110):
            params['use-exact-size'] = 'true'
        lun_create = netapp_api.NaElement.create_node_with_children(
            'lun-create-by-size',
            **params)
        if qos_policy_group_name:
            lun_create.add_new_child('qos-policy-group', qos_policy_group_name)

        try:
            self.connection.invoke_successfully(lun_create, True)
        except netapp_api.NaApiError as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Error provisioning volume %(lun_name)s on "
                          "%(volume_name)s. Details: %(ex)s",
                          {'lun_name': lun_name,
                           'volume_name': volume_name,
                           'ex': ex})

        if ontap_version < '9.5':
            self.do_direct_resize(path, six.text_type(size))
            if metadata['SpaceReserved'] == 'true':
                self.set_lun_space_reservation(path, True)

    def set_lun_space_reservation(self, path, flag):
        """Sets the LUN space reservation on ONTAP."""

        lun_modify_space_reservation = (
            netapp_api.NaElement.create_node_with_children(
                'lun-set-space-reservation-info', **{
                    'path': path,
                    'enable': str(flag)}))
        self.connection.invoke_successfully(lun_modify_space_reservation, True)

    def destroy_lun(self, path, force=True):
        """Destroys the LUN at the path."""
        lun_destroy = netapp_api.NaElement.create_node_with_children(
            'lun-destroy',
            **{'path': path})
        if force:
            lun_destroy.add_new_child('force', 'true')
        self.connection.invoke_successfully(lun_destroy, True)
        seg = path.split("/")
        LOG.debug("Destroyed LUN %s", seg[-1])

    def map_lun(self, path, igroup_name, lun_id=None):
        """Maps LUN to the initiator and returns LUN id assigned."""
        lun_map = netapp_api.NaElement.create_node_with_children(
            'lun-map', **{'path': path,
                          'initiator-group': igroup_name})
        if lun_id:
            lun_map.add_new_child('lun-id', lun_id)
        try:
            result = self.connection.invoke_successfully(lun_map, True)
            return result.get_child_content('lun-id-assigned')
        except netapp_api.NaApiError as e:
            code = e.code
            message = e.message
            LOG.warning('Error mapping LUN. Code :%(code)s, Message: '
                        '%(message)s', {'code': code, 'message': message})
            raise

    def unmap_lun(self, path, igroup_name):
        """Unmaps a LUN from given initiator."""
        lun_unmap = netapp_api.NaElement.create_node_with_children(
            'lun-unmap',
            **{'path': path, 'initiator-group': igroup_name})
        try:
            self.connection.invoke_successfully(lun_unmap, True)
        except netapp_api.NaApiError as e:
            exc_info = sys.exc_info()
            LOG.warning("Error unmapping LUN. Code :%(code)s, Message: "
                        "%(message)s", {'code': e.code,
                                        'message': e.message})
            # if the LUN is already unmapped
            if e.code == '13115' or e.code == '9016':
                pass
            else:
                six.reraise(*exc_info)

    def create_igroup(self, igroup, igroup_type='iscsi', os_type='default'):
        """Creates igroup with specified args."""
        igroup_create = netapp_api.NaElement.create_node_with_children(
            'igroup-create',
            **{'initiator-group-name': igroup,
               'initiator-group-type': igroup_type,
               'os-type': os_type})
        self.connection.invoke_successfully(igroup_create, True)

    def add_igroup_initiator(self, igroup, initiator):
        """Adds initiators to the specified igroup."""
        igroup_add = netapp_api.NaElement.create_node_with_children(
            'igroup-add',
            **{'initiator-group-name': igroup,
               'initiator': initiator})
        self.connection.invoke_successfully(igroup_add, True)

    def do_direct_resize(self, path, new_size_bytes, force=True):
        """Resize the LUN."""
        seg = path.split("/")
        LOG.info("Resizing LUN %s directly to new size.", seg[-1])
        lun_resize = netapp_api.NaElement.create_node_with_children(
            'lun-resize',
            **{'path': path,
               'size': new_size_bytes})
        if force:
            lun_resize.add_new_child('force', 'true')
        self.connection.invoke_successfully(lun_resize, True)

    def get_lun_geometry(self, path):
        """Gets the LUN geometry."""
        geometry = {}
        lun_geo = netapp_api.NaElement("lun-get-geometry")
        lun_geo.add_new_child('path', path)
        try:
            result = self.connection.invoke_successfully(lun_geo, True)
            geometry['size'] = result.get_child_content("size")
            geometry['bytes_per_sector'] = result.get_child_content(
                "bytes-per-sector")
            geometry['sectors_per_track'] = result.get_child_content(
                "sectors-per-track")
            geometry['tracks_per_cylinder'] = result.get_child_content(
                "tracks-per-cylinder")
            geometry['cylinders'] = result.get_child_content("cylinders")
            geometry['max_resize'] = result.get_child_content(
                "max-resize-size")
        except Exception as e:
            LOG.error("LUN %(path)s geometry failed. Message - %(msg)s",
                      {'path': path, 'msg': six.text_type(e)})
        return geometry

    def get_volume_options(self, volume_name):
        """Get the value for the volume option."""
        opts = []
        vol_option_list = netapp_api.NaElement("volume-options-list-info")
        vol_option_list.add_new_child('volume', volume_name)
        result = self.connection.invoke_successfully(vol_option_list, True)
        options = result.get_child_by_name("options")
        if options:
            opts = options.get_children()
        return opts

    def move_lun(self, path, new_path):
        """Moves the LUN at path to new path."""
        seg = path.split("/")
        new_seg = new_path.split("/")
        LOG.debug("Moving LUN %(name)s to %(new_name)s.",
                  {'name': seg[-1], 'new_name': new_seg[-1]})
        lun_move = netapp_api.NaElement("lun-move")
        lun_move.add_new_child("path", path)
        lun_move.add_new_child("new-path", new_path)
        self.connection.invoke_successfully(lun_move, True)

    def get_iscsi_target_details(self):
        """Gets the iSCSI target portal details."""
        raise NotImplementedError()

    def get_fc_target_wwpns(self):
        """Gets the FC target details."""
        raise NotImplementedError()

    def get_iscsi_service_details(self):
        """Returns iscsi iqn."""
        raise NotImplementedError()

    def check_iscsi_initiator_exists(self, iqn):
        """Returns True if initiator exists."""
        raise NotImplementedError()

    def set_iscsi_chap_authentication(self, iqn, username, password):
        """Provides NetApp host's CHAP credentials to the backend."""
        raise NotImplementedError()

    def get_lun_list(self):
        """Gets the list of LUNs on filer."""
        raise NotImplementedError()

    def get_igroup_by_initiators(self, initiator_list):
        """Get igroups exactly matching a set of initiators."""
        raise NotImplementedError()

    def _has_luns_mapped_to_initiator(self, initiator):
        """Checks whether any LUNs are mapped to the given initiator."""
        lun_list_api = netapp_api.NaElement('lun-initiator-list-map-info')
        lun_list_api.add_new_child('initiator', initiator)
        result = self.connection.invoke_successfully(lun_list_api, True)
        lun_maps_container = result.get_child_by_name(
            'lun-maps') or netapp_api.NaElement('none')
        return len(lun_maps_container.get_children()) > 0

    def has_luns_mapped_to_initiators(self, initiator_list):
        """Checks whether any LUNs are mapped to the given initiator(s)."""
        for initiator in initiator_list:
            if self._has_luns_mapped_to_initiator(initiator):
                return True
        return False

    def get_lun_by_args(self, **args):
        """Retrieves LUNs with specified args."""
        raise NotImplementedError()

    def get_performance_counter_info(self, object_name, counter_name):
        """Gets info about one or more Data ONTAP performance counters."""

        api_args = {'objectname': object_name}
        result = self.connection.send_request('perf-object-counter-list-info',
                                              api_args,
                                              enable_tunneling=False)

        counters = result.get_child_by_name(
            'counters') or netapp_api.NaElement('None')

        for counter in counters.get_children():

            if counter.get_child_content('name') == counter_name:

                labels = []
                label_list = counter.get_child_by_name(
                    'labels') or netapp_api.NaElement('None')
                for label in label_list.get_children():
                    labels.extend(label.get_content().split(','))
                base_counter = counter.get_child_content('base-counter')

                return {
                    'name': counter_name,
                    'labels': labels,
                    'base-counter': base_counter,
                }
        else:
            raise exception.NotFound(_('Counter %s not found') % counter_name)

    def delete_snapshot(self, volume_name, snapshot_name):
        """Deletes a volume snapshot."""
        api_args = {'volume': volume_name, 'snapshot': snapshot_name}
        self.connection.send_request('snapshot-delete', api_args)

    def create_cg_snapshot(self, volume_names, snapshot_name):
        """Creates a consistency group snapshot out of one or more flexvols.

        ONTAP requires an invocation of cg-start to first fence off the
        flexvols to be included in the snapshot. If cg-start returns
        success, a cg-commit must be executed to finalized the snapshot and
        unfence the flexvols.
        """
        cg_id = self._start_cg_snapshot(volume_names, snapshot_name)
        if not cg_id:
            msg = _('Could not start consistency group snapshot %s.')
            raise exception.VolumeBackendAPIException(data=msg % snapshot_name)
        self._commit_cg_snapshot(cg_id)

    def _start_cg_snapshot(self, volume_names, snapshot_name):
        snapshot_init = {
            'snapshot': snapshot_name,
            'timeout': 'relaxed',
            'volumes': [
                {'volume-name': volume_name} for volume_name in volume_names
            ],
        }
        result = self.connection.send_request('cg-start', snapshot_init)
        return result.get_child_content('cg-id')

    def _commit_cg_snapshot(self, cg_id):
        snapshot_commit = {'cg-id': cg_id}
        self.connection.send_request('cg-commit', snapshot_commit)

    def get_snapshot(self, volume_name, snapshot_name):
        """Gets a single snapshot."""
        raise NotImplementedError()

    @utils.retry(exception.SnapshotIsBusy)
    def wait_for_busy_snapshot(self, flexvol, snapshot_name):
        """Checks for and handles a busy snapshot.

        If a snapshot is busy, for reasons other than cloning, an exception is
        raised immediately. Otherwise, wait for a period of time for the clone
        dependency to finish before giving up. If the snapshot is not busy then
        no action is taken and the method exits.
        """
        snapshot = self.get_snapshot(flexvol, snapshot_name)
        if not snapshot['busy']:
            LOG.debug("Backing consistency group snapshot %s available for "
                      "deletion.", snapshot_name)
            return
        else:
            LOG.debug("Snapshot %(snap)s for vol %(vol)s is busy, waiting "
                      "for volume clone dependency to clear.",
                      {"snap": snapshot_name, "vol": flexvol})
            raise exception.SnapshotIsBusy(snapshot_name=snapshot_name)

    def mark_snapshot_for_deletion(self, volume, snapshot_name):
        """Mark snapshot for deletion by renaming snapshot."""
        return self.rename_snapshot(
            volume, snapshot_name, DELETED_PREFIX + snapshot_name)

    def rename_snapshot(self, volume, current_name, new_name):
        """Renames a snapshot."""
        api_args = {
            'volume': volume,
            'current-name': current_name,
            'new-name': new_name,
        }
        return self.connection.send_request('snapshot-rename', api_args)



# Copyright 2015 Intel
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect

import mock
import tooz.coordination
import tooz.locking

from cinder import coordination
from cinder import test

if hasattr(inspect, 'getfullargspec'):
    getargspec = inspect.getfullargspec
else:
    getargspec = inspect.getargspec


class Locked(Exception):
    pass


class MockToozLock(tooz.locking.Lock):
    active_locks = set()

    def acquire(self, blocking=True):
        if self.name not in self.active_locks:
            self.active_locks.add(self.name)
            return True
        elif not blocking:
            return False
        else:
            raise Locked

    def release(self):
        self.active_locks.remove(self.name)


@mock.patch('tooz.coordination.get_coordinator')
class CoordinatorTestCase(test.TestCase):
    MOCK_TOOZ = False

    def test_coordinator_start(self, get_coordinator):
        crd = get_coordinator.return_value

        agent = coordination.Coordinator()
        agent.start()
        self.assertTrue(get_coordinator.called)
        self.assertTrue(crd.start.called)

    def test_coordinator_stop(self, get_coordinator):
        crd = get_coordinator.return_value

        agent = coordination.Coordinator()
        agent.start()
        self.assertIsNotNone(agent.coordinator)
        agent.stop()
        self.assertTrue(crd.stop.called)
        self.assertIsNone(agent.coordinator)

    def test_coordinator_lock(self, get_coordinator):
        crd = get_coordinator.return_value
        crd.get_lock.side_effect = lambda n: MockToozLock(n)

        agent1 = coordination.Coordinator()
        agent1.start()
        agent2 = coordination.Coordinator()
        agent2.start()

        lock_name = 'lock'
        expected_name = lock_name.encode('ascii')

        self.assertNotIn(expected_name, MockToozLock.active_locks)
        with agent1.get_lock(lock_name):
            self.assertIn(expected_name, MockToozLock.active_locks)
            self.assertRaises(Locked, agent1.get_lock(lock_name).acquire)
            self.assertRaises(Locked, agent2.get_lock(lock_name).acquire)
        self.assertNotIn(expected_name, MockToozLock.active_locks)

    def test_coordinator_offline(self, get_coordinator):
        crd = get_coordinator.return_value
        crd.start.side_effect = tooz.coordination.ToozConnectionError('err')

        agent = coordination.Coordinator()
        self.assertRaises(tooz.coordination.ToozError, agent.start)
        self.assertFalse(agent.started)


@mock.patch.object(coordination.COORDINATOR, 'get_lock')
class CoordinationTestCase(test.TestCase):
    def test_synchronized(self, get_lock):
        @coordination.synchronized('lock-{f_name}-{foo.val}-{bar[val]}')
        def func(foo, bar):
            pass

        foo = mock.Mock()
        foo.val = 7
        bar = mock.MagicMock()
        bar.__getitem__.return_value = 8
        func(foo, bar)
        get_lock.assert_called_with('lock-func-7-8')
        self.assertEqual(['foo', 'bar'], getargspec(func)[0])



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Tests for volume init host method cases."""

import mock
from oslo_config import cfg
from oslo_utils import importutils

from cinder import context
from cinder import exception
from cinder import objects
from cinder.tests.unit import utils as tests_utils
from cinder.tests.unit import volume as base
from cinder.volume import driver
from cinder.volume import volume_migration as volume_migration
from cinder.volume import volume_utils


CONF = cfg.CONF


class VolumeInitHostTestCase(base.BaseVolumeTestCase):

    def setUp(self):
        super(VolumeInitHostTestCase, self).setUp()
        self.service_id = 1

    @mock.patch('cinder.manager.CleanableManager.init_host')
    def test_init_host_count_allocated_capacity(self, init_host_mock):
        vol0 = tests_utils.create_volume(
            self.context, size=100, host=CONF.host)
        vol1 = tests_utils.create_volume(
            self.context, size=128,
            host=volume_utils.append_host(CONF.host, 'pool0'))
        vol2 = tests_utils.create_volume(
            self.context, size=256,
            host=volume_utils.append_host(CONF.host, 'pool0'))
        vol3 = tests_utils.create_volume(
            self.context, size=512,
            host=volume_utils.append_host(CONF.host, 'pool1'))
        vol4 = tests_utils.create_volume(
            self.context, size=1024,
            host=volume_utils.append_host(CONF.host, 'pool2'))
        self.volume.init_host(service_id=self.service_id)
        init_host_mock.assert_called_once_with(
            service_id=self.service_id, added_to_cluster=None)
        stats = self.volume.stats
        self.assertEqual(2020, stats['allocated_capacity_gb'])
        self.assertEqual(
            384, stats['pools']['pool0']['allocated_capacity_gb'])
        self.assertEqual(
            512, stats['pools']['pool1']['allocated_capacity_gb'])
        self.assertEqual(
            1024, stats['pools']['pool2']['allocated_capacity_gb'])

        # NOTE(jdg): On the create we have host='xyz', BUT
        # here we do a db.volume_get, and now the host has
        # been updated to xyz#pool-name.  Note this is
        # done via the managers init, which calls the drivers
        # get_pool method, which in the legacy case is going
        # to be volume_backend_name or None

        vol0.refresh()
        expected_host = volume_utils.append_host(CONF.host, 'fake')
        self.assertEqual(expected_host, vol0.host)
        self.volume.delete_volume(self.context, vol0)
        self.volume.delete_volume(self.context, vol1)
        self.volume.delete_volume(self.context, vol2)
        self.volume.delete_volume(self.context, vol3)
        self.volume.delete_volume(self.context, vol4)

    def test_init_host_count_allocated_capacity_batch_retrieval(self):
        old_val = CONF.init_host_max_objects_retrieval
        CONF.init_host_max_objects_retrieval = 1
        try:
            self.test_init_host_count_allocated_capacity()
        finally:
            CONF.init_host_max_objects_retrieval = old_val

    @mock.patch('cinder.manager.CleanableManager.init_host')
    def test_init_host_count_allocated_capacity_cluster(self, init_host_mock):
        cluster_name = 'mycluster'
        self.volume.cluster = cluster_name
        # All these volumes belong to the same cluster, so we will calculate
        # the capacity of them all because we query the DB by cluster_name.
        tests_utils.create_volume(self.context, size=100, host=CONF.host,
                                  cluster_name=cluster_name)
        tests_utils.create_volume(
            self.context, size=128, cluster_name=cluster_name,
            host=volume_utils.append_host(CONF.host, 'pool0'))
        tests_utils.create_volume(
            self.context, size=256, cluster_name=cluster_name,
            host=volume_utils.append_host(CONF.host + '2', 'pool0'))
        tests_utils.create_volume(
            self.context, size=512, cluster_name=cluster_name,
            host=volume_utils.append_host(CONF.host + '2', 'pool1'))
        tests_utils.create_volume(
            self.context, size=1024, cluster_name=cluster_name,
            host=volume_utils.append_host(CONF.host + '3', 'pool2'))

        # These don't belong to the cluster so they will be ignored
        tests_utils.create_volume(
            self.context, size=1024,
            host=volume_utils.append_host(CONF.host, 'pool2'))
        tests_utils.create_volume(
            self.context, size=1024, cluster_name=cluster_name + '1',
            host=volume_utils.append_host(CONF.host + '3', 'pool2'))

        self.volume.init_host(service_id=self.service_id)
        init_host_mock.assert_called_once_with(
            service_id=self.service_id, added_to_cluster=None)
        stats = self.volume.stats
        self.assertEqual(2020, stats['allocated_capacity_gb'])
        self.assertEqual(
            384, stats['pools']['pool0']['allocated_capacity_gb'])
        self.assertEqual(
            512, stats['pools']['pool1']['allocated_capacity_gb'])
        self.assertEqual(
            1024, stats['pools']['pool2']['allocated_capacity_gb'])

    @mock.patch.object(driver.BaseVD, "update_provider_info")
    def test_init_host_sync_provider_info(self, mock_update):
        vol0 = tests_utils.create_volume(
            self.context, size=1, host=CONF.host)
        vol1 = tests_utils.create_volume(
            self.context, size=1, host=CONF.host)
        vol2 = tests_utils.create_volume(
            self.context, size=1, host=CONF.host, status='creating')
        snap0 = tests_utils.create_snapshot(self.context, vol0.id)
        snap1 = tests_utils.create_snapshot(self.context, vol1.id)
        # Return values for update_provider_info
        volumes = [{'id': vol0.id, 'provider_id': '1 2 xxxx'},
                   {'id': vol1.id, 'provider_id': '3 4 yyyy'}]
        snapshots = [{'id': snap0.id, 'provider_id': '5 6 xxxx'},
                     {'id': snap1.id, 'provider_id': '7 8 yyyy'}]
        mock_update.return_value = (volumes, snapshots)
        # initialize
        self.volume.init_host(service_id=self.service_id)
        # Grab volume and snapshot objects
        vol0_obj = objects.Volume.get_by_id(context.get_admin_context(),
                                            vol0.id)
        vol1_obj = objects.Volume.get_by_id(context.get_admin_context(),
                                            vol1.id)
        vol2_obj = objects.Volume.get_by_id(context.get_admin_context(),
                                            vol2.id)
        snap0_obj = objects.Snapshot.get_by_id(self.context, snap0.id)
        snap1_obj = objects.Snapshot.get_by_id(self.context, snap1.id)
        # Check updated provider ids
        self.assertEqual('1 2 xxxx', vol0_obj.provider_id)
        self.assertEqual('3 4 yyyy', vol1_obj.provider_id)
        self.assertIsNone(vol2_obj.provider_id)
        self.assertEqual('5 6 xxxx', snap0_obj.provider_id)
        self.assertEqual('7 8 yyyy', snap1_obj.provider_id)
        # Clean up
        self.volume.delete_snapshot(self.context, snap0_obj)
        self.volume.delete_snapshot(self.context, snap1_obj)
        self.volume.delete_volume(self.context, vol0)
        self.volume.delete_volume(self.context, vol1)

    def test_init_host_sync_provider_info_batch_retrieval(self):
        old_val = CONF.init_host_max_objects_retrieval
        CONF.init_host_max_objects_retrieval = 1
        try:
            self.test_init_host_sync_provider_info()
        finally:
            CONF.init_host_max_objects_retrieval = old_val

    @mock.patch.object(driver.BaseVD, "update_provider_info")
    def test_init_host_sync_provider_info_no_update(self, mock_update):
        vol0 = tests_utils.create_volume(
            self.context, size=1, host=CONF.host)
        vol1 = tests_utils.create_volume(
            self.context, size=1, host=CONF.host)
        snap0 = tests_utils.create_snapshot(self.context, vol0.id)
        snap1 = tests_utils.create_snapshot(self.context, vol1.id)
        mock_update.return_value = ([], [])
        # initialize
        self.volume.init_host(service_id=self.service_id)
        # Grab volume and snapshot objects
        vol0_obj = objects.Volume.get_by_id(context.get_admin_context(),
                                            vol0.id)
        vol1_obj = objects.Volume.get_by_id(context.get_admin_context(),
                                            vol1.id)
        snap0_obj = objects.Snapshot.get_by_id(self.context, snap0.id)
        snap1_obj = objects.Snapshot.get_by_id(self.context, snap1.id)
        # Check provider ids are not changed
        self.assertIsNone(vol0_obj.provider_id)
        self.assertIsNone(vol1_obj.provider_id)
        self.assertIsNone(snap0_obj.provider_id)
        self.assertIsNone(snap1_obj.provider_id)
        # Clean up
        self.volume.delete_snapshot(self.context, snap0_obj)
        self.volume.delete_snapshot(self.context, snap1_obj)
        self.volume.delete_volume(self.context, vol0)
        self.volume.delete_volume(self.context, vol1)

    @mock.patch.object(driver.BaseVD, "update_provider_info")
    def test_init_host_sync_provider_info_no_update_cluster(self, mock_update):
        cluster_name = 'mycluster'
        self.volume.cluster = cluster_name
        vol0 = tests_utils.create_volume(
            self.context, size=1, host=CONF.host, cluster_name=cluster_name)
        vol1 = tests_utils.create_volume(
            self.context, size=1, host=CONF.host + '2',
            cluster_name=cluster_name)
        vol2 = tests_utils.create_volume(
            self.context, size=1, host=CONF.host)
        vol3 = tests_utils.create_volume(
            self.context, size=1, host=CONF.host,
            cluster_name=cluster_name + '2')
        snap0 = tests_utils.create_snapshot(self.context, vol0.id)
        snap1 = tests_utils.create_snapshot(self.context, vol1.id)
        tests_utils.create_snapshot(self.context, vol2.id)
        tests_utils.create_snapshot(self.context, vol3.id)
        mock_update.return_value = ([], [])
        # initialize
        self.volume.init_host(service_id=self.service_id)
        # Grab volume and snapshot objects
        vol0_obj = objects.Volume.get_by_id(context.get_admin_context(),
                                            vol0.id)
        vol1_obj = objects.Volume.get_by_id(context.get_admin_context(),
                                            vol1.id)
        snap0_obj = objects.Snapshot.get_by_id(self.context, snap0.id)
        snap1_obj = objects.Snapshot.get_by_id(self.context, snap1.id)

        self.assertSetEqual({vol0.id, vol1.id},
                            {vol.id for vol in mock_update.call_args[0][0]})
        self.assertSetEqual({snap0.id, snap1.id},
                            {snap.id for snap in mock_update.call_args[0][1]})
        # Check provider ids are not changed
        self.assertIsNone(vol0_obj.provider_id)
        self.assertIsNone(vol1_obj.provider_id)
        self.assertIsNone(snap0_obj.provider_id)
        self.assertIsNone(snap1_obj.provider_id)
        # Clean up
        self.volume.delete_snapshot(self.context, snap0_obj)
        self.volume.delete_snapshot(self.context, snap1_obj)
        self.volume.delete_volume(self.context, vol0)
        self.volume.delete_volume(self.context, vol1)

    @mock.patch('cinder.volume.manager.VolumeManager.'
                '_include_resources_in_cluster')
    def test_init_host_cluster_not_changed(self, include_in_cluster_mock):
        self.volume.init_host(added_to_cluster=False,
                              service_id=self.service_id)
        include_in_cluster_mock.assert_not_called()

    @mock.patch('cinder.objects.group.GroupList.include_in_cluster')
    @mock.patch('cinder.objects.snapshot.SnapshotList.get_all',
                return_value=[])
    @mock.patch('cinder.objects.volume.VolumeList.get_all', return_value=[])
    @mock.patch('cinder.objects.volume.VolumeList.include_in_cluster')
    @mock.patch('cinder.objects.consistencygroup.ConsistencyGroupList.'
                'include_in_cluster')
    @mock.patch('cinder.db.image_volume_cache_include_in_cluster')
    def test_init_host_added_to_cluster(self, image_cache_include_mock,
                                        cg_include_mock,
                                        vol_include_mock, vol_get_all_mock,
                                        snap_get_all_mock, group_include_mock):
        cluster = str(mock.sentinel.cluster)
        self.mock_object(self.volume, 'cluster', cluster)
        self.volume.init_host(added_to_cluster=True,
                              service_id=self.service_id)

        vol_include_mock.assert_called_once_with(mock.ANY, cluster,
                                                 host=self.volume.host)
        cg_include_mock.assert_called_once_with(mock.ANY, cluster,
                                                host=self.volume.host)
        image_cache_include_mock.assert_called_once_with(mock.ANY, cluster,
                                                         host=self.volume.host)
        group_include_mock.assert_called_once_with(mock.ANY, cluster,
                                                   host=self.volume.host)
        vol_get_all_mock.assert_called_once_with(
            mock.ANY, filters={'cluster_name': cluster},
            limit=None, offset=None)
        snap_get_all_mock.assert_called_once_with(
            mock.ANY, filters={'cluster_name': cluster},
            limit=None, offset=None)

    @mock.patch('cinder.keymgr.migration.migrate_fixed_key')
    @mock.patch('cinder.volume.manager.VolumeManager._get_my_volumes')
    @mock.patch('cinder.manager.ThreadPoolManager._add_to_threadpool')
    def test_init_host_key_migration(self,
                                     mock_add_threadpool,
                                     mock_get_my_volumes,
                                     mock_migrate_fixed_key):

        self.volume.init_host(service_id=self.service_id)

        volumes = mock_get_my_volumes()
        volumes_to_migrate = volume_migration.VolumeMigrationList()
        volumes_to_migrate.append(volumes, self.context)
        mock_add_threadpool.assert_called_once_with(
            mock_migrate_fixed_key,
            volumes=volumes_to_migrate)

    @mock.patch('time.sleep')
    def test_init_host_retry(self, mock_sleep):
        kwargs = {'service_id': 2}
        self.volume = importutils.import_object(CONF.volume_manager)
        self.volume.driver.do_setup = mock.MagicMock()
        self.volume.driver.do_setup.side_effect = [
            exception.CinderException("Test driver error."),
            exception.InvalidConfigurationValue('Test config error.'),
            ImportError]

        self.volume.init_host(added_to_cluster=False, **kwargs)

        self.assertEqual(4, self.volume.driver.do_setup.call_count)
        self.assertFalse(self.volume.is_working())

    @mock.patch('time.sleep')
    def test_init_host_retry_once(self, mock_sleep):
        kwargs = {'service_id': 2}
        self.volume = importutils.import_object(CONF.volume_manager)
        self.volume.driver.do_setup = mock.MagicMock()
        self.volume.driver.do_setup.side_effect = [ImportError, None]

        self.volume.init_host(added_to_cluster=False, **kwargs)

        self.assertEqual(2, self.volume.driver.do_setup.call_count)
        self.assertTrue(self.volume.is_working())



# Copyright 2017 Datera
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import six

from cinder import context
from cinder import exception
from cinder import test
from cinder.volume import configuration as conf
from cinder.volume.drivers.datera import datera_common as datc
from cinder.volume.drivers.datera import datera_iscsi as datera
from cinder.volume import volume_types


datc.DEFAULT_SI_SLEEP = 0
datc.DEFAULT_SI_SLEEP_API_2 = 0
datc.DEFAULT_SNAP_SLEEP = 0
URL_TEMPLATES = datera.datc.URL_TEMPLATES
OS_PREFIX = datera.datc.OS_PREFIX
UNMANAGE_PREFIX = datera.datc.UNMANAGE_PREFIX


class DateraVolumeTestCasev2(test.TestCase):

    def setUp(self):
        super(DateraVolumeTestCasev2, self).setUp()

        self.cfg = mock.Mock(spec=conf.Configuration)
        self.cfg.san_ip = '127.0.0.1'
        self.cfg.san_is_local = True
        self.cfg.datera_api_token = 'secret'
        self.cfg.datera_api_port = '7717'
        self.cfg.datera_num_replicas = '2'
        self.cfg.datera_503_timeout = 0.01
        self.cfg.datera_503_interval = 0.001
        self.cfg.datera_acl_allow_all = False
        self.cfg.datera_debug = False
        self.cfg.san_login = 'user'
        self.cfg.san_password = 'pass'
        self.cfg.datera_tenant_id = 'test-tenant'
        self.cfg.driver_client_cert = None
        self.cfg.driver_client_cert_key = None
        self.cfg.datera_disable_profiler = False
        self.cfg.driver_use_ssl = False

        mock_exec = mock.Mock()
        mock_exec.return_value = ('', '')

        self.driver = datera.DateraDriver(execute=mock_exec,
                                          configuration=self.cfg)
        self.driver.set_initialized()
        self.driver.configuration.get = _config_getter
        self.volume = _stub_volume()
        self.driver._request = mock.Mock()
        m = mock.Mock()
        m.json.return_value = {'api_versions': ['v2']}
        self.driver._request.return_value = m
        self.mock_api = mock.Mock()
        self.driver._issue_api_request = self.mock_api
        self._apiv = "2"
        self._tenant = None

        # self.addCleanup(self.api_patcher.stop)

    def test_volume_create_success(self):
        self.mock_api.return_value = stub_single_ai
        self.assertIsNone(self.driver.create_volume(self.volume))

    def test_volume_create_fails(self):
        self.mock_api.side_effect = datc.DateraAPIException
        self.assertRaises(datc.DateraAPIException,
                          self.driver.create_volume, self.volume)

    def test_volume_create_delay(self):
        """Verify after 1st retry volume becoming available is a success."""

        def _progress_api_return(mock_api):
            if mock_api.retry_count == 1:
                _bad_vol_ai = stub_single_ai.copy()
                _bad_vol_ai['storage_instances']['storage-1'][
                    'volumes']['volume-1']['op_status'] = 'unavailable'
                return _bad_vol_ai
            else:
                self.mock_api.retry_count += 1
                return stub_single_ai
        self.mock_api.retry_count = 0
        self.mock_api.return_value = _progress_api_return(self.mock_api)
        self.assertEqual(1, self.mock_api.retry_count)
        self.assertIsNone(self.driver.create_volume(self.volume))

    @mock.patch.object(volume_types, 'get_volume_type')
    def test_create_volume_with_extra_specs(self, mock_get_type):
        self.mock_api.return_value = stub_single_ai
        mock_get_type.return_value = {
            'name': u'The Best',
            'qos_specs_id': None,
            'deleted': False,
            'created_at': '2015-08-14 04:18:11',
            'updated_at': None,
            'extra_specs': {
                u'volume_backend_name': u'datera',
                u'qos:max_iops_read': u'2000',
                u'qos:max_iops_write': u'4000',
                u'qos:max_iops_total': u'4000'
            },
            'is_public': True,
            'deleted_at': None,
            'id': u'dffb4a83-b8fb-4c19-9f8c-713bb75db3b1',
            'description': None
        }

        mock_volume = _stub_volume(
            volume_type_id='dffb4a83-b8fb-4c19-9f8c-713bb75db3b1'
        )

        self.assertIsNone(self.driver.create_volume(mock_volume))
        self.assertTrue(mock_get_type.called)

    def test_create_cloned_volume_success(self):
        source_volume = _stub_volume(
            id='7f91abfa-7964-41ed-88fc-207c3a290b4f',
            display_name='foo'
        )
        self.assertIsNone(self.driver.create_cloned_volume(self.volume,
                                                           source_volume))

    def test_create_cloned_volume_success_larger(self):
        cloned_volume = _stub_volume(
            id='7f91abfa-7964-41ed-88fc-207c3a290b4f',
            display_name='foo',
            size=2
        )

        mock_extend = mock.Mock()
        if self._apiv == '2':
            self.driver._extend_volume_2 = mock_extend
            self.driver.create_cloned_volume(cloned_volume, self.volume)
            mock_extend.assert_called_with(
                cloned_volume, cloned_volume['size'])
        else:
            self.driver._extend_volume_2_1 = mock_extend
            self.driver.create_cloned_volume(cloned_volume, self.volume)
            mock_extend.assert_called_with(
                cloned_volume, cloned_volume['size'])

    def test_create_cloned_volume_fails(self):
        self.mock_api.side_effect = datc.DateraAPIException
        source_volume = _stub_volume(
            id='7f91abfa-7964-41ed-88fc-207c3a290b4f',
            display_name='foo'
        )
        self.assertRaises(datc.DateraAPIException,
                          self.driver.create_cloned_volume, self.volume,
                          source_volume)

    def test_delete_volume_success(self):
        if self._apiv == '2':
            self.mock_api.side_effect = [
                {},
                self._generate_fake_api_request()(
                    "acl_policy", api_version=self._apiv, tenant=self._tenant),
                self._generate_fake_api_request()(
                    "ig_group", api_version=self._apiv, tenant=self._tenant),
                {},
                {},
                {},
                {},
                {}]
        else:
            self.mock_api.side_effect = [
                {},
                {},
                self._generate_fake_api_request()(
                    "acl_policy", api_version=self._apiv, tenant=self._tenant),
                self._generate_fake_api_request()(
                    "ig_group", api_version=self._apiv, tenant=self._tenant),
                {},
                {},
                {},
                {},
                {}]
        self.assertIsNone(self.driver.delete_volume(self.volume))

    def test_delete_volume_not_found(self):
        if self._apiv == '2':
            self.mock_api.side_effect = exception.NotFound
        else:
            self.mock_api.side_effect = [
                self._generate_fake_api_request("tenant"),
                {},
                exception.NotFound,
                self._generate_fake_api_request()(
                    "acl_policy", api_version=self._apiv, tenant=self._tenant),
                self._generate_fake_api_request()(
                    "ig_group", api_version=self._apiv, tenant=self._tenant),
                {},
                {},
                {},
                {},
            ]
        self.assertIsNone(self.driver.delete_volume(self.volume))

    def test_delete_volume_fails(self):
        self.mock_api.side_effect = datc.DateraAPIException
        self.assertRaises(datc.DateraAPIException,
                          self.driver.delete_volume, self.volume)

    def test_ensure_export_success(self):
        self.mock_api.side_effect = self._generate_fake_api_request()
        ctxt = context.get_admin_context()
        self.assertIsNone(self.driver.ensure_export(ctxt,
                                                    self.volume,
                                                    None))

    def test_ensure_export_fails(self):
        self.mock_api.side_effect = datc.DateraAPIException
        ctxt = context.get_admin_context()
        self.assertRaises(datc.DateraAPIException,
                          self.driver.ensure_export, ctxt, self.volume, None)

    def test_create_export_target_does_not_exist_success(self):
        self.mock_api.side_effect = self._generate_fake_api_request(
            targets_exist=False)
        ctxt = context.get_admin_context()
        self.assertIsNone(self.driver.create_export(ctxt,
                                                    self.volume,
                                                    None))

    def test_create_export_fails(self):
        self.mock_api.side_effect = datc.DateraAPIException
        ctxt = context.get_admin_context()
        self.assertRaises(datc.DateraAPIException,
                          self.driver.create_export,
                          ctxt,
                          self.volume,
                          None)

    def test_initialize_connection_success(self):
        self.mock_api.side_effect = self._generate_fake_api_request()
        connector = {}

        expected = {
            'driver_volume_type': 'iscsi',
            'data': {
                'target_discovered': False,
                'volume_id': self.volume['id'],
                'target_iqn': ('iqn.2013-05.com.daterainc:tc:01:sn:'
                               '3bbb080aab7d9abc'),
                'target_portal': '172.28.41.63:3260',
                'target_lun': 0,
                'discard': False}}
        self.assertEqual(expected,
                         self.driver.initialize_connection(self.volume,
                                                           connector))

    def test_initialize_connection_fails(self):
        self.mock_api.side_effect = datc.DateraAPIException
        connector = {}
        self.assertRaises(datc.DateraAPIException,
                          self.driver.initialize_connection,
                          self.volume,
                          connector)

    def test_detach_volume_success(self):
        if self._apiv == '2':
            self.mock_api.side_effect = [
                {},
                self._generate_fake_api_request()(
                    "acl_policy", api_version=self._apiv, tenant=self._tenant),
                self._generate_fake_api_request()(
                    "ig_group", api_version=self._apiv, tenant=self._tenant),
                {},
                {},
                {},
                {}]
        else:
            self.mock_api.side_effect = [
                {},
                {},
                self._generate_fake_api_request()(
                    "acl_policy", api_version=self._apiv, tenant=self._tenant),
                self._generate_fake_api_request()(
                    "ig_group", api_version=self._apiv, tenant=self._tenant),
                {},
                {},
                {},
                {}]
        ctxt = context.get_admin_context()
        volume = _stub_volume(status='in-use')
        self.assertIsNone(self.driver.detach_volume(ctxt, volume))

    def test_detach_volume_fails(self):
        self.mock_api.side_effect = datc.DateraAPIException
        ctxt = context.get_admin_context()
        volume = _stub_volume(status='in-use')
        self.assertRaises(datc.DateraAPIException,
                          self.driver.detach_volume, ctxt, volume)

    def test_detach_volume_not_found(self):
        if self._apiv == '2':
            self.mock_api.side_effect = exception.NotFound
        else:
            self.mock_api.side_effect = [
                self._generate_fake_api_request("tenant"),
                exception.NotFound,
                self._generate_fake_api_request()(
                    "acl_policy", api_version=self._apiv, tenant=self._tenant),
                self._generate_fake_api_request()(
                    "ig_group", api_version=self._apiv, tenant=self._tenant),
                {},
                {},
                {},
                {}]
        ctxt = context.get_admin_context()
        volume = _stub_volume(status='in-use')
        self.assertIsNone(self.driver.detach_volume(ctxt, volume))

    def test_create_snapshot_success(self):
        snapshot = _stub_snapshot(volume_id=self.volume['id'])
        self.mock_api.side_effect = self._generate_fake_api_request()
        self.assertIsNone(self.driver.create_snapshot(snapshot))

    def test_create_snapshot_fails(self):
        self.mock_api.side_effect = datc.DateraAPIException
        snapshot = _stub_snapshot(volume_id=self.volume['id'])
        self.assertRaises(datc.DateraAPIException,
                          self.driver.create_snapshot, snapshot)

    def test_delete_snapshot_success(self):
        if self._apiv == '2':
            self.mock_api.return_value = stub_return_snapshots
        else:
            self.mock_api.return_value = stub_return_snapshots_21
        snapshot = _stub_snapshot(volume_id=self.volume['id'])
        self.assertIsNone(self.driver.delete_snapshot(snapshot))

    def test_delete_snapshot_not_found(self):
        if self._apiv == '2':
            self.mock_api.side_effect = [
                stub_return_snapshots,
                exception.NotFound]
        else:
            self.mock_api.side_effect = [
                self._generate_fake_api_request("tenant"),
                stub_return_snapshots_21,
                exception.NotFound]
        snapshot = _stub_snapshot(self.volume['id'], volume_id="test")
        self.assertIsNone(self.driver.delete_snapshot(snapshot))

    def test_delete_snapshot_fails(self):
        self.mock_api.side_effect = datc.DateraAPIException
        snapshot = _stub_snapshot(volume_id=self.volume['id'])
        self.assertRaises(datc.DateraAPIException,
                          self.driver.delete_snapshot, snapshot)

    def test_create_volume_from_snapshot_success(self):
        snapshot = _stub_snapshot(volume_id=self.volume['id'])
        if self._apiv == '2':
            self.mock_api.side_effect = [
                stub_return_snapshots,
                list(stub_return_snapshots.values())[0],
                None]
        else:
            self.mock_api.side_effect = [
                self._generate_fake_api_request("tenant"),
                stub_return_snapshots_21,
                {'data': stub_return_snapshots_21['data'][0]},
                None]
        self.assertIsNone(
            self.driver.create_volume_from_snapshot(self.volume, snapshot))

    @mock.patch.object(datera.DateraDriver, 'extend_volume')
    def test_create_volume_from_snapshot_success_larger(self, mock_extend):
        snapshot = _stub_snapshot(volume_id=self.volume['id'])
        extend_volume = _stub_volume(size=2)

        mock_extend = mock.Mock()
        if self._apiv == '2':
            self.driver._extend_volume_2 = mock_extend
            self.mock_api.side_effect = [
                stub_return_snapshots,
                list(stub_return_snapshots.values())[0],
                None]
            self.driver.create_volume_from_snapshot(extend_volume, snapshot)
            mock_extend.assert_called_once_with(extend_volume,
                                                extend_volume['size'])
        else:
            self.driver._extend_volume_2_1 = mock_extend
            self.mock_api.side_effect = [
                self._generate_fake_api_request("tenant"),
                stub_return_snapshots_21,
                {'data': stub_return_snapshots_21['data'][0]},
                None]
            self.driver.create_volume_from_snapshot(extend_volume, snapshot)
            mock_extend.assert_called_once_with(extend_volume,
                                                extend_volume['size'])

    def test_create_volume_from_snapshot_fails(self):
        self.mock_api.side_effect = datc.DateraAPIException
        snapshot = _stub_snapshot(volume_id=self.volume['id'])
        self.assertRaises(datc.DateraAPIException,
                          self.driver.create_volume_from_snapshot, self.volume,
                          snapshot)

    def test_extend_volume_success(self):
        volume = _stub_volume(size=1)
        self.mock_api.side_effect = [
            stub_get_export,
            {'data': stub_get_export},
            self._generate_fake_api_request()(
                "acl_policy", api_version=self._apiv, tenant=self._tenant),
            self._generate_fake_api_request()(
                "ig_group", api_version=self._apiv, tenant=self._tenant),
            self._generate_fake_api_request()(
                "acl_policy", api_version=self._apiv, tenant=self._tenant),
            {}, {}, {}, {}, {}, {}, stub_get_export,
            {'data': stub_get_export}]
        self.assertIsNone(self.driver.extend_volume(volume, 2))

    def test_extend_volume_fails(self):
        self.mock_api.side_effect = datc.DateraAPIException
        volume = _stub_volume(size=1)
        self.assertRaises(datc.DateraAPIException,
                          self.driver.extend_volume, volume, 2)

    def test_login_successful(self):
        self.mock_api.return_value = {
            'key': 'dd2469de081346c28ac100e071709403'
        }
        self.assertIsNone(self.driver.login())
        self.assertEqual(1, self.mock_api.call_count)

    def test_login_unsuccessful(self):
        self.mock_api.side_effect = exception.NotAuthorized
        self.assertRaises(exception.NotAuthorized, self.driver.login)
        self.assertEqual(1, self.mock_api.call_count)

    def test_manage_existing(self):
        self.mock_api.return_value = {}
        if self._apiv == '2':
            test_name = {"source-name": "test-app:test-si:test-vol"}
            self.assertIsNone(
                self.driver.manage_existing(
                    _stub_volume(),
                    test_name))
            self.mock_api.assert_called_with(
                URL_TEMPLATES['ai_inst']().format(
                    test_name["source-name"].split(":")[0]),
                method='put',
                body={'name': OS_PREFIX + _stub_volume()['id']},
                api_version=self._apiv)
        else:
            tenant = 'tenant'
            test_name = {"source-name": "{}:test-app:test-si:test-vol".format(
                tenant)}
            self.assertIsNone(
                self.driver.manage_existing(
                    _stub_volume(),
                    test_name))
            self.mock_api.assert_called_with(
                URL_TEMPLATES['ai_inst']().format(
                    test_name["source-name"].split(":")[1]),
                method='put',
                body={'name': OS_PREFIX + _stub_volume()['id']},
                api_version=self._apiv,
                tenant='tenant')

    def test_manage_existing_wrong_ref(self):
        TEST_NAME = {"source-name": "incorrect-reference"}
        self.assertRaises(
            exception.ManageExistingInvalidReference,
            self.driver.manage_existing,
            _stub_volume(),
            TEST_NAME)

    def test_manage_existing_get_size(self):
        TEST_NAME = {"source-name": "test-app:storage-1:volume-1"}
        self.mock_api.side_effect = self._generate_fake_api_request()
        self.assertEqual(
            self.driver.manage_existing_get_size(
                _stub_volume(),
                TEST_NAME), 500)
        if self._apiv == '2':
            self.mock_api.assert_called_with(
                URL_TEMPLATES['ai_inst']().format(
                    TEST_NAME["source-name"].split(":")[0]),
                api_version=self._apiv)
        else:
            self.mock_api.assert_called_with(
                URL_TEMPLATES['ai_inst']().format(
                    TEST_NAME["source-name"].split(":")[0]),
                api_version=self._apiv,
                tenant=self._tenant)

    def test_manage_existing_get_size_wrong_ref(self):
        TEST_NAME = {"source-name": "incorrect-reference"}
        self.assertRaises(
            exception.ManageExistingInvalidReference,
            self.driver.manage_existing_get_size,
            _stub_volume(),
            TEST_NAME)

    def test_get_manageable_volumes(self):
        if self._apiv == '2':
            self.mock_api.return_value = non_cinder_ais
            six.assertCountEqual(
                self,
                self.driver.get_manageable_volumes(
                    {}, "", 10, 0, "", ""),
                [{'cinder_id': None,
                  'extra_info': None,
                  'reason_not_safe': None,
                  'reference': {
                      "source-name": 'test-app-inst:storage-1:volume-1'},
                  'safe_to_manage': True,
                  'size': 50},
                 {'cinder_id': 'c20aba21-6ef6-446b-b374-45733b4883ba',
                  'extra_info': None,
                  'reason_not_safe': None,
                  'reference': None,
                  'safe_to_manage': False,
                  'size': None}])
        else:
            self.mock_api.return_value = non_cinder_ais_21
            self.assertEqual(
                self.driver.get_manageable_volumes(
                    {}, "", 10, 0, "", ""),
                [{'cinder_id': None,
                  'extra_info': None,
                  'reason_not_safe': '',
                  'reference': {
                      "source-name": 'test-app-inst:storage-1:volume-1'},
                  'safe_to_manage': True,
                  'size': 50},
                 {'cinder_id': 'c20aba21-6ef6-446b-b374-45733b4883ba',
                  'extra_info': None,
                  'reason_not_safe': '',
                  'reference': None,
                  'safe_to_manage': False,
                  'size': None}])

    def test_unmanage(self):
        self.mock_api.return_value = {}
        self.assertIsNone(self.driver.unmanage(_stub_volume()))
        if self._apiv == '2':
            self.mock_api.assert_called_with(
                URL_TEMPLATES['ai_inst']().format(
                    OS_PREFIX + _stub_volume()['id']),
                method='put',
                body={'name': UNMANAGE_PREFIX + _stub_volume()['id']},
                api_version=self._apiv)
        else:
            self.mock_api.assert_called_with(
                URL_TEMPLATES['ai_inst']().format(
                    OS_PREFIX + _stub_volume()['id']),
                method='put',
                body={'name': UNMANAGE_PREFIX + _stub_volume()['id']},
                api_version=self._apiv,
                tenant=self._tenant)

    def _generate_fake_api_request(self, targets_exist=True):
        def _fake_api_request(resource_type, *args, **kwargs):
            if 'api_version' not in kwargs:
                raise ValueError("Fix me dummy")
            result = None
            if resource_type.split('/')[-1] == 'storage-1':
                result = stub_get_export
            elif (resource_type.split('/')[-1] ==
                    'c20aba21-6ef6-446b-b374-45733b4883ba'):
                result = stub_app_instance[
                    'c20aba21-6ef6-446b-b374-45733b4883ba']
            elif resource_type == 'acl_policy':
                result = stub_acl if self._apiv == '2' else stub_acl_21
            elif resource_type == 'ig_group':
                result = stub_ig if self._apiv == '2' else stub_ig_21
            elif resource_type.split('/')[-1] == 'snapshots':
                result = {'timestamp': 'test_ts'}
            elif resource_type.split('/')[-1] == 'test_ts':
                result = {'op_state': 'available'}
            elif resource_type == 'tenant':
                return {}
            else:
                if kwargs.get('api_version') == '2':
                    result = list(stub_app_instance.values())[0]
                else:
                    result = stub_app_instance_21['data']

            if kwargs.get('api_version') == '2':
                return result
            else:
                return {'data': result}
        return _fake_api_request


class DateraVolumeTestCasev21(DateraVolumeTestCasev2):

    def setUp(self):
        super(DateraVolumeTestCasev21, self).setUp()

        m = mock.Mock()
        m.json.return_value = {'api_versions': ['v2.1']}
        self.driver._request.return_value = m
        self.driver._store_metadata = mock.Mock()
        self._apiv = '2.1'
        self._tenant = self.cfg.datera_tenant_id


stub_acl = {
    'initiator_groups': [
        '/initiator_groups/IG-8739f309-dae9-4534-aa02-5b8e9e96eefd'],
    'initiators': [],
    'path': ('/app_instances/8739f309-dae9-4534-aa02-5b8e9e96eefd/'
             'storage_instances/storage-1/acl_policy')}

stub_acl_21 = {
    'initiator_groups': [
        {'path': '/initiator_groups/IG-8739f309-dae9-4534-aa02-5b8e9e96eefd'}],
    'initiators': [],
    'path': ('/app_instances/8739f309-dae9-4534-aa02-5b8e9e96eefd/'
             'storage_instances/storage-1/acl_policy')}

stub_ig = {
    'members': ['/initiators/iqn.1993-08.org.debian:01:ed22de8d75c0'],
    'name': 'IG-21e08155-8b95-4108-b148-089f64623963',
    'path': '/initiator_groups/IG-21e08155-8b95-4108-b148-089f64623963'}

stub_ig_21 = {
    'members': [
        {'path': '/initiators/iqn.1993-08.org.debian:01:ed22de8d75c0'}],
    'name': 'IG-21e08155-8b95-4108-b148-089f64623963',
    'path': '/initiator_groups/IG-21e08155-8b95-4108-b148-089f64623963'}

stub_create_export = {
    "_ipColl": ["172.28.121.10", "172.28.120.10"],
    "acls": {},
    "activeServers": {"4594953e-f97f-e111-ad85-001e6738c0f0": "1"},
    "ctype": "TC_BLOCK_ISCSI",
    "endpointsExt1": {
        "4594953e-f97f-e111-ad85-001e6738c0f0": {
            "ipHigh": 0,
            "ipLow": "192421036",
            "ipStr": "172.28.120.11",
            "ipV": 4,
            "name": "",
            "network": 24
        }
    },
    "endpointsExt2": {
        "4594953e-f97f-e111-ad85-001e6738c0f0": {
            "ipHigh": 0,
            "ipLow": "192486572",
            "ipStr": "172.28.121.11",
            "ipV": 4,
            "name": "",
            "network": 24
        }
    },
    "inodes": {"c20aba21-6ef6-446b-b374-45733b4883ba": "1"},
    "name": "",
    "networkPort": 0,
    "serverAllocation": "TS_ALLOC_COMPLETED",
    "servers": {"4594953e-f97f-e111-ad85-001e6738c0f0": "1"},
    "targetAllocation": "TS_ALLOC_COMPLETED",
    "targetIds": {
        "4594953e-f97f-e111-ad85-001e6738c0f0": {
            "ids": [{
                "dev": None,
                "id": "iqn.2013-05.com.daterainc::01:sn:fc372bc0490b2dbe"
            }]
        }
    },
    "typeName": "TargetIscsiConfig",
    "uuid": "7071efd7-9f22-4996-8f68-47e9ab19d0fd"
}


stub_app_instance = {
    "c20aba21-6ef6-446b-b374-45733b4883ba": {
        "admin_state": "online",
        "clone_src": {},
        "create_mode": "openstack",
        "descr": "",
        "health": "ok",
        "name": "c20aba21-6ef6-446b-b374-45733b4883ba",
        "path": "/app_instances/c20aba21-6ef6-446b-b374-45733b4883ba",
        "storage_instances": {
            "storage-1": {
                "access": {
                    "ips": [
                        "172.28.41.63"
                    ],
                    "iqn": "iqn.2013-05.com.daterainc:tc:01:sn:"
                           "3bbb080aab7d9abc",
                    "path": "/app_instances/c20aba21-6ef6-446b-b374"
                            "-45733b4883ba/storage_instances/storage-1/access"
                },
                "access_control": {
                    "initiator_groups": [],
                    "initiators": [],
                    "path": "/app_instances/c20aba21-6ef6-446b-b374-"
                            "45733b4883ba/storage_instances/storage-1"
                            "/access_control"
                },
                "access_control_mode": "allow_all",
                "active_initiators": [],
                "active_storage_nodes": [
                    "/storage_nodes/1c4feac4-17c7-478b-8928-c76e8ec80b72"
                ],
                "admin_state": "online",
                "auth": {
                    "initiator_pswd": "",
                    "initiator_user_name": "",
                    "path": "/app_instances/c20aba21-6ef6-446b-b374-"
                            "45733b4883ba/storage_instances/storage-1/auth",
                    "target_pswd": "",
                    "target_user_name": "",
                    "type": "none"
                },
                "creation_type": "user",
                "descr": "c20aba21-6ef6-446b-b374-45733b4883ba__ST__storage-1",
                "op_state": "available",
                "name": "storage-1",
                "path": "/app_instances/c20aba21-6ef6-446b-b374-"
                        "45733b4883ba/storage_instances/storage-1",
                "uuid": "b9897b84-149f-43c7-b19c-27d6af8fa815",
                "volumes": {
                    "volume-1": {
                        "capacity_in_use": 0,
                        "name": "volume-1",
                        "op_state": "available",
                        "path": "/app_instances/c20aba21-6ef6-446b-b374-"
                                "45733b4883ba/storage_instances/storage-1"
                                "/volumes/volume-1",
                        "replica_count": 3,
                        "size": 500,
                        "snapshot_policies": {},
                        "snapshots": {
                            "1445384931.322468627": {
                                "op_state": "available",
                                "path": "/app_instances/c20aba21-6ef6-446b"
                                        "-b374-45733b4883ba/storage_instances"
                                        "/storage-1/volumes/volume-1/snapshots"
                                        "/1445384931.322468627",
                                "uuid": "0bb34f0c-fea4-48e0-bf96-591120ac7e3c"
                            }
                        },
                        "uuid": "c20aba21-6ef6-446b-b374-45733b4883ba"
                    }
                }
            }
        },
        "uuid": "c20aba21-6ef6-446b-b374-45733b4883ba"
    }
}
stub_app_instance_21 = {
    "tenant": "/root",
    "path": "/app_instances/1e52946a-5c77-45ed-8b4e-b46e7236a8eb",
    "version": "v2.1",
    "data": {
        "tenant": "/root",
        "path": "/app_instances/1e52946a-5c77-45ed-8b4e-b46e7236a8eb",
        "name": "OS-9b0216bc-8aab-47f2-b746-843f497cb7a6",
        "id": "1e52946a-5c77-45ed-8b4e-b46e7236a8eb",
        "health": "ok",
        "app_template": {
            "path": "",
            "resolved_path": "",
            "resolved_tenant": ""
        },
        "descr": "",
        "admin_state": "online",
        "storage_instances": [
            {
                "health": "ok",
                "path": "/app_instances/1e52946a-5c77-45ed-8b4e-b46e7236a8eb/"
                "storage_instances/storage-1",
                "name": "storage-1",
                "admin_state": "online",
                "op_state": "available",
                "volumes": [
                    {
                        "path": "/app_instances/1e52946a-5c77-45ed-8b4e-"
                                "b46e7236a8eb/"
                        "storage_instances/storage-1/volumes/volume-1",
                        "name": "volume-1",
                        "replica_count": 1,
                        "uuid": "9b0216bc-8aab-47f2-b746-843f497cb7a6",
                        "size": 500,
                        "capacity_in_use": 0,
                        "snapshot_policies": [],
                        "snapshots": [],
                        "placement_mode": "hybrid",
                        "op_state": "available",
                        "active_storage_nodes": [
                            {
                                "path": "/storage_nodes/75f2cae4-68fb-4236-"
                                        "a90c-b6c480b68816"
                            }
                        ],
                        "health": "ok"
                    }
                ],
                "access_control_mode": "deny_all",
                "acl_policy": {
                    "path": "/app_instances/1e52946a-5c77-45ed-8b4e-"
                            "b46e7236a8eb/"
                    "storage_instances/storage-1/acl_policy",
                    "initiators": [],
                    "initiator_groups": []
                },
                "ip_pool": {
                    "path": "/access_network_ip_pools/default",
                    "resolved_path": "/access_network_ip_pools/default",
                    "resolved_tenant": "/root"
                },
                "access": {
                    "path": "/app_instances/1e52946a-5c77-45ed-8b4e-"
                            "b46e7236a8eb/"
                    "storage_instances/storage-1/access",
                    "ips": [
                        "172.28.41.63",
                        "172.29.41.29"
                    ],
                    "iqn": "iqn.2013-05.com.daterainc:tc:01:sn:"
                           "3bbb080aab7d9abc"
                },
                "auth": {
                    "path": "/app_instances/1e52946a-5c77-45ed-8b4e-"
                            "b46e7236a8eb/"
                    "storage_instances/storage-1/auth",
                    "type": "none",
                    "initiator_user_name": "",
                    "initiator_pswd": "(hidden)",
                    "target_user_name": "",
                    "target_pswd": "(hidden)"
                },
                "active_initiators": [],
                "active_storage_nodes": [
                    {
                        "path": "/storage_nodes/75f2cae4-68fb-4236-a90c-"
                                "b6c480b68816"
                    }
                ],
                "uuid": "eb3d7b07-b520-4cc2-b365-90135b84c356"
            }
        ],
        "create_mode": "openstack",
        "uuid": "9b0216bc-8aab-47f2-b746-843f497cb7a6",
        "snapshots": [],
        "snapshot_policies": []
    }
}

stub_get_export = stub_app_instance[
    'c20aba21-6ef6-446b-b374-45733b4883ba']['storage_instances']['storage-1']

stub_single_ai = stub_app_instance['c20aba21-6ef6-446b-b374-45733b4883ba']

stub_return_snapshots = \
    {
        "1446076293.118600738": {
            "op_state": "available",
            "path": "/app_instances/c20aba21-6ef6-446b-b374-45733b4883ba"
            "/storage_instances/storage-1/volumes/volume-1/snapshots/"
            "1446076293.118600738",
            "uuid": "0bb34f0c-fea4-48e0-bf96-591120ac7e3c"
        },
        "1446076384.00607846": {
            "op_state": "available",
            "path": "/app_instances/c20aba21-6ef6-446b-b374-45733b4883ba"
            "/storage_instances/storage-1/volumes/volume-1/snapshots/"
            "1446076384.00607846",
            "uuid": "25b4b959-c30a-45f2-a90c-84a40f34f0a1"
        }
    }

stub_return_snapshots_21 = {
    'data': [
        {
            "op_state": "available",
            "path": "/app_instances/c20aba21-6ef6-446b-b374-45733b4883ba"
            "/storage_instances/storage-1/volumes/volume-1/snapshots/",
            "timestamp": "1446076293.118600738",
            "utc_ts": "1446076293.118600738",
            "uuid": "0bb34f0c-fea4-48e0-bf96-591120ac7e3c"
        },
        {
            "op_state": "available",
            "path": "/app_instances/c20aba21-6ef6-446b-b374-45733b4883ba"
            "/storage_instances/storage-1/volumes/volume-1/snapshots/",
            "timestamp": "1446076384.00607846",
            "utc_ts": "1446076384.00607846",
            "uuid": "25b4b959-c30a-45f2-a90c-84a40f34f0a1"
        }]
}

non_cinder_ais = {
    "75bc1c69-a399-4acb-aade-3514caf13c5e": {
        "admin_state": "online",
        "create_mode": "normal",
        "descr": "",
        "health": "ok",
        "id": "75bc1c69-a399-4acb-aade-3514caf13c5e",
        "name": "test-app-inst",
        "path": "/app_instances/75bc1c69-a399-4acb-aade-3514caf13c5e",
        "snapshot_policies": {},
        "snapshots": {},
        "storage_instances": {
            "storage-1": {
                "access": {
                    "ips": [
                        "172.28.41.93"
                    ],
                    "iqn": "iqn.2013-05.com.daterainc:tc:01:sn:"
                           "29036682e2d37b98",
                    "path": "/app_instances/75bc1c69-a399-4acb-aade-"
                            "3514caf13c5e/storage_instances/storage-1/access"
                },
                "access_control_mode": "deny_all",
                "acl_policy": {
                    "initiator_groups": [],
                    "initiators": [],
                    "path": "/app_instances/75bc1c69-a399-4acb-aade-"
                            "3514caf13c5e/storage_instances/storage-"
                            "1/acl_policy"
                },
                "active_initiators": [],
                "active_storage_nodes": [
                    "/storage_nodes/78b350a8-43f2-453f-a257-8df76d7406b9"
                ],
                "admin_state": "online",
                "auth": {
                    "initiator_pswd": "(hidden)",
                    "initiator_user_name": "",
                    "path": "/app_instances/75bc1c69-a399-4acb-aade-"
                            "3514caf13c5e/storage_instances/storage-1/auth",
                    "target_pswd": "(hidden)",
                    "target_user_name": "",
                    "type": "none"
                },
                "creation_type": "user",
                "ip_pool": "/access_network_ip_pools/default",
                "name": "storage-1",
                "op_state": "available",
                "path": "/app_instances/75bc1c69-a399-4acb-aade-"
                        "3514caf13c5e/storage_instances/storage-1",
                "uuid": "6421237d-e4fc-433a-b535-148d5b6d8586",
                "volumes": {
                    "volume-1": {
                        "capacity_in_use": 0,
                        "name": "volume-1",
                        "op_state": "available",
                        "path": "/app_instances/75bc1c69-a399-4acb-aade-"
                                "3514caf13c5e/storage_instances/storage-"
                                "1/volumes/volume-1",
                        "replica_count": 1,
                        "size": 50,
                        "snapshot_policies": {},
                        "snapshots": {},
                        "uuid": "e674d29c-a672-40d1-9577-abe3a504ffe9"
                    }
                }
            }
        },
        "uuid": "00000000-0000-0000-0000-000000000000"
    },
    "dfdaf8d1-8976-4c13-a829-3345e03cf810": {
        "admin_state": "offline",
        "create_mode": "openstack",
        "descr": "",
        "health": "ok",
        "id": "dfdaf8d1-8976-4c13-a829-3345e03cf810",
        "name": "OS-c20aba21-6ef6-446b-b374-45733b4883ba",
        "path": "/app_instances/dfdaf8d1-8976-4c13-a829-3345e03cf810",
        "snapshot_policies": {},
        "snapshots": {},
        "storage_instances": {
            "storage-1": {
                "access": {
                    "ips": [
                        "172.28.41.57"
                    ],
                    "iqn": "iqn.2013-05.com.daterainc:tc:01:sn:"
                           "56cd59e754ad02b6",
                    "path": "/app_instances/dfdaf8d1-8976-4c13-a829-"
                            "3345e03cf810/storage_instances/storage-1/access"
                },
                "access_control_mode": "deny_all",
                "acl_policy": {
                    "initiator_groups": [],
                    "initiators": [],
                    "path": "/app_instances/dfdaf8d1-8976-4c13-a829-"
                            "3345e03cf810/storage_instances/storage-"
                            "1/acl_policy"
                },
                "active_initiators": [],
                "active_storage_nodes": [
                    "/storage_nodes/78b350a8-43f2-453f-a257-8df76d7406b9"
                ],
                "admin_state": "offline",
                "auth": {
                    "initiator_pswd": "(hidden)",
                    "initiator_user_name": "",
                    "path": "/app_instances/dfdaf8d1-8976-4c13-a829-"
                            "3345e03cf810/storage_instances/storage-1/auth",
                    "target_pswd": "(hidden)",
                    "target_user_name": "",
                    "type": "none"
                },
                "creation_type": "user",
                "ip_pool": "/access_network_ip_pools/default",
                "name": "storage-1",
                "op_state": "unavailable",
                "path": "/app_instances/dfdaf8d1-8976-4c13-a829-3345e03cf810"
                        "/storage_instances/storage-1",
                "uuid": "5620a673-9985-464e-9616-e325a50eac60",
                "volumes": {
                    "volume-1": {
                        "capacity_in_use": 0,
                        "name": "volume-1",
                        "op_state": "available",
                        "path": "/app_instances/dfdaf8d1-8976-4c13-a829-"
                                "3345e03cf810/storage_instances/storage-"
                                "1/volumes/volume-1",
                        "replica_count": 1,
                        "size": 5,
                        "snapshot_policies": {},
                        "snapshots": {},
                        "uuid": "c20aba21-6ef6-446b-b374-45733b4883ba"
                    }
                }
            }
        },
        "uuid": "c20aba21-6ef6-446b-b374-45733b4883ba"
    }
}

non_cinder_ais_21 = {
    'data': [{
        "admin_state": "online",
        "create_mode": "normal",
        "descr": "",
        "health": "ok",
        "id": "75bc1c69-a399-4acb-aade-3514caf13c5e",
        "name": "test-app-inst",
        "path": "/app_instances/75bc1c69-a399-4acb-aade-3514caf13c5e",
        "snapshot_policies": {},
        "snapshots": {},
        "storage_instances": {
            "storage-1": {
                "access": {
                    "ips": [
                        "172.28.41.93"
                    ],
                    "iqn": "iqn.2013-05.com.daterainc:tc:01:sn:"
                           "29036682e2d37b98",
                    "path": "/app_instances/75bc1c69-a399-4acb-aade-"
                            "3514caf13c5e/storage_instances/storage-1/access"
                },
                "access_control_mode": "deny_all",
                "acl_policy": {
                    "initiator_groups": [],
                    "initiators": [],
                    "path": "/app_instances/75bc1c69-a399-4acb-aade-"
                            "3514caf13c5e/storage_instances/storage-"
                            "1/acl_policy"
                },
                "active_initiators": [],
                "active_storage_nodes": [
                    "/storage_nodes/78b350a8-43f2-453f-a257-8df76d7406b9"
                ],
                "admin_state": "online",
                "auth": {
                    "initiator_pswd": "(hidden)",
                    "initiator_user_name": "",
                    "path": "/app_instances/75bc1c69-a399-4acb-aade-"
                            "3514caf13c5e/storage_instances/storage-1/auth",
                    "target_pswd": "(hidden)",
                    "target_user_name": "",
                    "type": "none"
                },
                "creation_type": "user",
                "ip_pool": "/access_network_ip_pools/default",
                "name": "storage-1",
                "op_state": "available",
                "path": "/app_instances/75bc1c69-a399-4acb-aade-"
                        "3514caf13c5e/storage_instances/storage-1",
                "uuid": "6421237d-e4fc-433a-b535-148d5b6d8586",
                "volumes": {
                    "volume-1": {
                        "capacity_in_use": 0,
                        "name": "volume-1",
                        "op_state": "available",
                        "path": "/app_instances/75bc1c69-a399-4acb-aade-"
                                "3514caf13c5e/storage_instances/storage-"
                                "1/volumes/volume-1",
                        "replica_count": 1,
                        "size": 50,
                        "snapshot_policies": {},
                        "snapshots": {},
                        "uuid": "e674d29c-a672-40d1-9577-abe3a504ffe9"
                    }
                }
            }
        },
        "uuid": "00000000-0000-0000-0000-000000000000"
    },
        {
        "admin_state": "offline",
        "create_mode": "openstack",
        "descr": "",
        "health": "ok",
        "id": "dfdaf8d1-8976-4c13-a829-3345e03cf810",
        "name": "OS-c20aba21-6ef6-446b-b374-45733b4883ba",
        "path": "/app_instances/dfdaf8d1-8976-4c13-a829-3345e03cf810",
        "snapshot_policies": {},
        "snapshots": {},
        "storage_instances": {
            "storage-1": {
                "access": {
                    "ips": [
                        "172.28.41.57"
                    ],
                    "iqn": "iqn.2013-05.com.daterainc:tc:01:sn:"
                           "56cd59e754ad02b6",
                    "path": "/app_instances/dfdaf8d1-8976-4c13-a829-"
                            "3345e03cf810/storage_instances/storage-1/access"
                },
                "access_control_mode": "deny_all",
                "acl_policy": {
                    "initiator_groups": [],
                    "initiators": [],
                    "path": "/app_instances/dfdaf8d1-8976-4c13-a829-"
                            "3345e03cf810/storage_instances/storage-"
                            "1/acl_policy"
                },
                "active_initiators": [],
                "active_storage_nodes": [
                    "/storage_nodes/78b350a8-43f2-453f-a257-8df76d7406b9"
                ],
                "admin_state": "offline",
                "auth": {
                    "initiator_pswd": "(hidden)",
                    "initiator_user_name": "",
                    "path": "/app_instances/dfdaf8d1-8976-4c13-a829-"
                            "3345e03cf810/storage_instances/storage-1/auth",
                    "target_pswd": "(hidden)",
                    "target_user_name": "",
                    "type": "none"
                },
                "creation_type": "user",
                "ip_pool": "/access_network_ip_pools/default",
                "name": "storage-1",
                "op_state": "unavailable",
                "path": "/app_instances/dfdaf8d1-8976-4c13-a829-3345e03cf810"
                        "/storage_instances/storage-1",
                "uuid": "5620a673-9985-464e-9616-e325a50eac60",
                "volumes": {
                    "volume-1": {
                        "capacity_in_use": 0,
                        "name": "volume-1",
                        "op_state": "available",
                        "path": "/app_instances/dfdaf8d1-8976-4c13-a829-"
                                "3345e03cf810/storage_instances/storage-"
                                "1/volumes/volume-1",
                        "replica_count": 1,
                        "size": 5,
                        "snapshot_policies": {},
                        "snapshots": {},
                        "uuid": "c20aba21-6ef6-446b-b374-45733b4883ba"
                    }
                }
            }
        },
        "uuid": "c20aba21-6ef6-446b-b374-45733b4883ba"
    }]
}


def _stub_datera_volume(*args, **kwargs):
    return {
        "status": "available",
        "name": "test",
        "num_replicas": "2",
        "parent": "00000000-0000-0000-0000-000000000000",
        "size": "1024",
        "sub_type": "IS_ORIGINAL",
        "uuid": "10305aa4-1343-4363-86fe-f49eb421a48c",
        "snapshots": [],
        "snapshot_configs": [],
        "targets": [
            kwargs.get('targets', "744e1bd8-d741-4919-86cd-806037d98c8a"),
        ]
    }


def _stub_volume(*args, **kwargs):
    uuid = u'c20aba21-6ef6-446b-b374-45733b4883ba'
    name = u'volume-00000001'
    size = 1
    volume = {}
    volume['id'] = kwargs.get('id', uuid)
    volume['display_name'] = kwargs.get('display_name', name)
    volume['size'] = kwargs.get('size', size)
    volume['provider_location'] = kwargs.get('provider_location', None)
    volume['volume_type_id'] = kwargs.get('volume_type_id', None)
    return volume


def _stub_snapshot(*args, **kwargs):
    uuid = u'0bb34f0c-fea4-48e0-bf96-591120ac7e3c'
    name = u'snapshot-00000001'
    size = 1
    volume = {}
    volume['id'] = kwargs.get('id', uuid)
    volume['display_name'] = kwargs.get('display_name', name)
    volume['volume_size'] = kwargs.get('size', size)
    volume['volume_id'] = kwargs.get('volume_id', None)
    return volume


def _config_getter(*args, **kwargs):
    return {}



# Copyright (c) 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Manage backends in the current zone."""

# TODO(smcginnis) update this once six has support for collections.abc
# (https://github.com/benjaminp/six/pull/241) or clean up once we drop py2.7.
try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils
from oslo_utils import strutils
from oslo_utils import timeutils

from cinder.common import constants
from cinder import context as cinder_context
from cinder import exception
from cinder import objects
from cinder.scheduler import filters
from cinder import utils
from cinder.volume import volume_types
from cinder.volume import volume_utils


# FIXME: This file should be renamed to backend_manager, we should also rename
# HostManager class, and scheduler_host_manager option, and also the weight
# classes, and add code to maintain backward compatibility.


host_manager_opts = [
    cfg.ListOpt('scheduler_default_filters',
                default=[
                    'AvailabilityZoneFilter',
                    'CapacityFilter',
                    'CapabilitiesFilter'
                ],
                help='Which filter class names to use for filtering hosts '
                     'when not specified in the request.'),
    cfg.ListOpt('scheduler_default_weighers',
                default=[
                    'CapacityWeigher'
                ],
                help='Which weigher class names to use for weighing hosts.'),
    cfg.StrOpt('scheduler_weight_handler',
               default='cinder.scheduler.weights.OrderedHostWeightHandler',
               help='Which handler to use for selecting the host/pool '
                    'after weighing'),
]

CONF = cfg.CONF
CONF.register_opts(host_manager_opts)
CONF.import_opt('scheduler_driver', 'cinder.scheduler.manager')
CONF.import_opt('max_over_subscription_ratio', 'cinder.volume.driver')

LOG = logging.getLogger(__name__)


class ReadOnlyDict(Mapping):
    """A read-only dict."""
    def __init__(self, source=None):
        if source is not None:
            self.data = dict(source)
        else:
            self.data = {}

    def __getitem__(self, key):
        return self.data[key]

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.data)


class BackendState(object):
    """Mutable and immutable information tracked for a volume backend."""

    def __init__(self, host, cluster_name, capabilities=None, service=None):
        # NOTE(geguileo): We have a circular dependency between BackendState
        # and PoolState and we resolve it with an instance attribute instead
        # of a class attribute that we would assign after the PoolState
        # declaration because this way we avoid splitting the code.
        self.pool_state_cls = PoolState

        self.capabilities = None
        self.service = None
        self.host = host
        self.cluster_name = cluster_name
        self.update_capabilities(capabilities, service)

        self.volume_backend_name = None
        self.vendor_name = None
        self.driver_version = 0
        self.storage_protocol = None
        self.QoS_support = False
        # Mutable available resources.
        # These will change as resources are virtually "consumed".
        self.total_capacity_gb = 0
        # capacity has been allocated in cinder POV, which should be
        # sum(vol['size'] for vol in vols_on_hosts)
        self.allocated_capacity_gb = 0
        self.free_capacity_gb = None
        self.reserved_percentage = 0
        # The apparent allocated space indicating how much capacity
        # has been provisioned. This could be the sum of sizes of
        # all volumes on a backend, which could be greater than or
        # equal to the allocated_capacity_gb.
        self.provisioned_capacity_gb = 0
        self.max_over_subscription_ratio = 1.0
        self.thin_provisioning_support = False
        self.thick_provisioning_support = False
        # Does this backend support attaching a volume to more than
        # once host/instance?
        self.multiattach = False

        # PoolState for all pools
        self.pools = {}

        self.updated = None

    @property
    def backend_id(self):
        return self.cluster_name or self.host

    def update_capabilities(self, capabilities=None, service=None):
        # Read-only capability dicts

        if capabilities is None:
            capabilities = {}
        self.capabilities = ReadOnlyDict(capabilities)
        if service is None:
            service = {}
        self.service = ReadOnlyDict(service)

    def update_from_volume_capability(self, capability, service=None):
        """Update information about a host from its volume_node info.

        'capability' is the status info reported by volume backend, a typical
        capability looks like this:

        .. code-block:: python

         {
          capability = {
              'volume_backend_name': 'Local iSCSI', #
              'vendor_name': 'OpenStack',           #  backend level
              'driver_version': '1.0',              #  mandatory/fixed
              'storage_protocol': 'iSCSI',          #  stats&capabilities

              'active_volumes': 10,                 #
              'IOPS_provisioned': 30000,            #  optional custom
              'fancy_capability_1': 'eat',          #  stats & capabilities
              'fancy_capability_2': 'drink',        #

              'pools': [
                  {'pool_name': '1st pool',         #
                   'total_capacity_gb': 500,        #  mandatory stats for
                   'free_capacity_gb': 230,         #  pools
                   'allocated_capacity_gb': 270,    #
                   'QoS_support': 'False',          #
                   'reserved_percentage': 0,        #

                   'dying_disks': 100,              #
                   'super_hero_1': 'spider-man',    #  optional custom
                   'super_hero_2': 'flash',         #  stats & capabilities
                   'super_hero_3': 'neoncat'        #
                  },
                  {'pool_name': '2nd pool',
                   'total_capacity_gb': 1024,
                   'free_capacity_gb': 1024,
                   'allocated_capacity_gb': 0,
                   'QoS_support': 'False',
                   'reserved_percentage': 0,

                   'dying_disks': 200,
                   'super_hero_1': 'superman',
                   'super_hero_2': ' ',
                   'super_hero_2': 'Hulk'
                  }
              ]
          }
         }

        """
        self.update_capabilities(capability, service)

        if capability:
            if self.updated and self.updated > capability['timestamp']:
                return

            # Update backend level info
            self.update_backend(capability)

            # Update pool level info
            self.update_pools(capability, service)

    def update_pools(self, capability, service):
        """Update storage pools information from backend reported info."""
        if not capability:
            return

        pools = capability.get('pools', None)
        active_pools = set()
        if pools and isinstance(pools, list):
            # Update all pools stats according to information from list
            # of pools in volume capacity
            for pool_cap in pools:
                pool_name = pool_cap['pool_name']
                self._append_backend_info(pool_cap)
                cur_pool = self.pools.get(pool_name, None)
                if not cur_pool:
                    # Add new pool
                    cur_pool = self.pool_state_cls(self.host,
                                                   self.cluster_name,
                                                   pool_cap,
                                                   pool_name)
                    self.pools[pool_name] = cur_pool
                cur_pool.update_from_volume_capability(pool_cap, service)

                active_pools.add(pool_name)
        elif pools is None:
            # To handle legacy driver that doesn't report pool
            # information in the capability, we have to prepare
            # a pool from backend level info, or to update the one
            # we created in self.pools.
            pool_name = self.volume_backend_name
            if pool_name is None:
                # To get DEFAULT_POOL_NAME
                pool_name = volume_utils.extract_host(self.host, 'pool', True)

            if len(self.pools) == 0:
                # No pool was there
                single_pool = self.pool_state_cls(self.host, self.cluster_name,
                                                  capability, pool_name)
                self._append_backend_info(capability)
                self.pools[pool_name] = single_pool
            else:
                # this is an update from legacy driver
                try:
                    single_pool = self.pools[pool_name]
                except KeyError:
                    single_pool = self.pool_state_cls(self.host,
                                                      self.cluster_name,
                                                      capability,
                                                      pool_name)
                    self._append_backend_info(capability)
                    self.pools[pool_name] = single_pool

            single_pool.update_from_volume_capability(capability, service)
            active_pools.add(pool_name)

        # remove non-active pools from self.pools
        nonactive_pools = set(self.pools.keys()) - active_pools
        for pool in nonactive_pools:
            LOG.debug("Removing non-active pool %(pool)s @ %(host)s "
                      "from scheduler cache.", {'pool': pool,
                                                'host': self.host})
            del self.pools[pool]

    def _append_backend_info(self, pool_cap):
        # Fill backend level info to pool if needed.
        if not pool_cap.get('volume_backend_name', None):
            pool_cap['volume_backend_name'] = self.volume_backend_name

        if not pool_cap.get('storage_protocol', None):
            pool_cap['storage_protocol'] = self.storage_protocol

        if not pool_cap.get('vendor_name', None):
            pool_cap['vendor_name'] = self.vendor_name

        if not pool_cap.get('driver_version', None):
            pool_cap['driver_version'] = self.driver_version

        if not pool_cap.get('timestamp', None):
            pool_cap['timestamp'] = self.updated

    def update_backend(self, capability):
        self.volume_backend_name = capability.get('volume_backend_name', None)
        self.vendor_name = capability.get('vendor_name', None)
        self.driver_version = capability.get('driver_version', None)
        self.storage_protocol = capability.get('storage_protocol', None)
        self.updated = capability['timestamp']

    def consume_from_volume(self, volume, update_time=True):
        """Incrementally update host state from a volume."""
        volume_gb = volume['size']
        self.allocated_capacity_gb += volume_gb
        self.provisioned_capacity_gb += volume_gb
        if self.free_capacity_gb == 'infinite':
            # There's virtually infinite space on back-end
            pass
        elif self.free_capacity_gb == 'unknown':
            # Unable to determine the actual free space on back-end
            pass
        else:
            self.free_capacity_gb -= volume_gb
        if update_time:
            self.updated = timeutils.utcnow()
        LOG.debug("Consumed %s GB from backend: %s", volume['size'], self)

    def __repr__(self):
        # FIXME(zhiteng) backend level free_capacity_gb isn't as
        # meaningful as it used to be before pool is introduced, we'd
        # come up with better representation of HostState.
        grouping = 'cluster' if self.cluster_name else 'host'
        grouping_name = self.backend_id
        return ("%(grouping)s '%(grouping_name)s':"
                "free_capacity_gb: %(free_capacity_gb)s, "
                "total_capacity_gb: %(total_capacity_gb)s,"
                "allocated_capacity_gb: %(allocated_capacity_gb)s, "
                "max_over_subscription_ratio: %(mosr)s,"
                "reserved_percentage: %(reserved_percentage)s, "
                "provisioned_capacity_gb: %(provisioned_capacity_gb)s,"
                "thin_provisioning_support: %(thin_provisioning_support)s, "
                "thick_provisioning_support: %(thick)s,"
                "pools: %(pools)s,"
                "updated at: %(updated)s" %
                {'grouping': grouping, 'grouping_name': grouping_name,
                 'free_capacity_gb': self.free_capacity_gb,
                 'total_capacity_gb': self.total_capacity_gb,
                 'allocated_capacity_gb': self.allocated_capacity_gb,
                 'mosr': self.max_over_subscription_ratio,
                 'reserved_percentage': self.reserved_percentage,
                 'provisioned_capacity_gb': self.provisioned_capacity_gb,
                 'thin_provisioning_support': self.thin_provisioning_support,
                 'thick': self.thick_provisioning_support,
                 'pools': self.pools, 'updated': self.updated})


class PoolState(BackendState):
    def __init__(self, host, cluster_name, capabilities, pool_name):
        new_host = volume_utils.append_host(host, pool_name)
        new_cluster = volume_utils.append_host(cluster_name, pool_name)
        super(PoolState, self).__init__(new_host, new_cluster, capabilities)
        self.pool_name = pool_name
        # No pools in pool
        self.pools = None

    def update_from_volume_capability(self, capability, service=None):
        """Update information about a pool from its volume_node info."""
        LOG.debug("Updating capabilities for %s: %s", self.host, capability)
        self.update_capabilities(capability, service)
        if capability:
            if self.updated and self.updated > capability['timestamp']:
                return
            self.update_backend(capability)

            self.total_capacity_gb = capability.get('total_capacity_gb', 0)
            self.free_capacity_gb = capability.get('free_capacity_gb', 0)
            self.allocated_capacity_gb = capability.get(
                'allocated_capacity_gb', 0)
            self.QoS_support = capability.get('QoS_support', False)
            self.reserved_percentage = capability.get('reserved_percentage', 0)
            # provisioned_capacity_gb is the apparent total capacity of
            # all the volumes created on a backend, which is greater than
            # or equal to allocated_capacity_gb, which is the apparent
            # total capacity of all the volumes created on a backend
            # in Cinder. Using allocated_capacity_gb as the default of
            # provisioned_capacity_gb if it is not set.
            self.provisioned_capacity_gb = capability.get(
                'provisioned_capacity_gb', self.allocated_capacity_gb)
            self.thin_provisioning_support = capability.get(
                'thin_provisioning_support', False)
            self.thick_provisioning_support = capability.get(
                'thick_provisioning_support', False)

            self.max_over_subscription_ratio = (
                utils.calculate_max_over_subscription_ratio(
                    capability, CONF.max_over_subscription_ratio))

            self.multiattach = capability.get('multiattach', False)

    def update_pools(self, capability):
        # Do nothing, since we don't have pools within pool, yet
        pass


class HostManager(object):
    """Base HostManager class."""

    backend_state_cls = BackendState

    REQUIRED_KEYS = frozenset([
        'pool_name',
        'total_capacity_gb',
        'free_capacity_gb',
        'allocated_capacity_gb',
        'provisioned_capacity_gb',
        'thin_provisioning_support',
        'thick_provisioning_support',
        'max_over_subscription_ratio',
        'reserved_percentage'])

    def __init__(self):
        self.service_states = {}  # { <host|cluster>: {<service>: {cap k : v}}}
        self.backend_state_map = {}
        self.filter_handler = filters.BackendFilterHandler('cinder.scheduler.'
                                                           'filters')
        self.filter_classes = self.filter_handler.get_all_classes()
        self.enabled_filters = self._choose_backend_filters(
            CONF.scheduler_default_filters)
        self.weight_handler = importutils.import_object(
            CONF.scheduler_weight_handler,
            'cinder.scheduler.weights')
        self.weight_classes = self.weight_handler.get_all_classes()

        self._no_capabilities_backends = set()  # Services without capabilities
        self._update_backend_state_map(cinder_context.get_admin_context())
        self.service_states_last_update = {}

    def _choose_backend_filters(self, filter_cls_names):
        """Return a list of available filter names.

        This function checks input filter names against a predefined set
        of acceptable filters (all loaded filters). If input is None,
        it uses CONF.scheduler_default_filters instead.
        """
        if not isinstance(filter_cls_names, (list, tuple)):
            filter_cls_names = [filter_cls_names]
        good_filters = []
        bad_filters = []
        for filter_name in filter_cls_names:
            found_class = False
            for cls in self.filter_classes:
                if cls.__name__ == filter_name:
                    found_class = True
                    good_filters.append(cls)
                    break
            if not found_class:
                bad_filters.append(filter_name)
        if bad_filters:
            raise exception.SchedulerHostFilterNotFound(
                filter_name=", ".join(bad_filters))
        return good_filters

    def _choose_backend_weighers(self, weight_cls_names):
        """Return a list of available weigher names.

        This function checks input weigher names against a predefined set
        of acceptable weighers (all loaded weighers).  If input is None,
        it uses CONF.scheduler_default_weighers instead.
        """
        if weight_cls_names is None:
            weight_cls_names = CONF.scheduler_default_weighers
        if not isinstance(weight_cls_names, (list, tuple)):
            weight_cls_names = [weight_cls_names]

        good_weighers = []
        bad_weighers = []
        for weigher_name in weight_cls_names:
            found_class = False
            for cls in self.weight_classes:
                if cls.__name__ == weigher_name:
                    good_weighers.append(cls)
                    found_class = True
                    break
            if not found_class:
                bad_weighers.append(weigher_name)
        if bad_weighers:
            raise exception.SchedulerHostWeigherNotFound(
                weigher_name=", ".join(bad_weighers))
        return good_weighers

    def get_filtered_backends(self, backends, filter_properties,
                              filter_class_names=None):
        """Filter backends and return only ones passing all filters."""
        if filter_class_names is not None:
            filter_classes = self._choose_backend_filters(filter_class_names)
        else:
            filter_classes = self.enabled_filters
        return self.filter_handler.get_filtered_objects(filter_classes,
                                                        backends,
                                                        filter_properties)

    def get_weighed_backends(self, backends, weight_properties,
                             weigher_class_names=None):
        """Weigh the backends."""
        weigher_classes = self._choose_backend_weighers(weigher_class_names)

        weighed_backends = self.weight_handler.get_weighed_objects(
            weigher_classes, backends, weight_properties)

        LOG.debug("Weighed %s", weighed_backends)
        return weighed_backends

    def update_service_capabilities(self, service_name, host, capabilities,
                                    cluster_name, timestamp):
        """Update the per-service capabilities based on this notification."""
        if service_name != 'volume':
            LOG.debug('Ignoring %(service_name)s service update '
                      'from %(host)s',
                      {'service_name': service_name, 'host': host})
            return

        # Determine whether HostManager has just completed initialization, and
        # has not received the rpc message returned by volume.
        just_init = self._is_just_initialized()

        # TODO(geguileo): In P - Remove the next line since we receive the
        # timestamp
        timestamp = timestamp or timeutils.utcnow()
        # Copy the capabilities, so we don't modify the original dict
        capab_copy = dict(capabilities)
        capab_copy["timestamp"] = timestamp

        # Set the default capabilities in case None is set.
        backend = cluster_name or host
        capab_old = self.service_states.get(backend, {"timestamp": 0})
        capab_last_update = self.service_states_last_update.get(
            backend, {"timestamp": 0})

        # Ignore older updates
        if capab_old['timestamp'] and timestamp < capab_old['timestamp']:
            LOG.info('Ignoring old capability report from %s.', backend)
            return

        # If the capabilities are not changed and the timestamp is older,
        # record the capabilities.

        # There are cases: capab_old has the capabilities set,
        # but the timestamp may be None in it. So does capab_last_update.

        if (not self._get_updated_pools(capab_old, capab_copy)) and (
                (not capab_old.get("timestamp")) or
                (not capab_last_update.get("timestamp")) or
                (capab_last_update["timestamp"] < capab_old["timestamp"])):
            self.service_states_last_update[backend] = capab_old

        self.service_states[backend] = capab_copy

        cluster_msg = (('Cluster: %s - Host: ' % cluster_name) if cluster_name
                       else '')
        LOG.debug("Received %(service_name)s service update from %(cluster)s "
                  "%(host)s: %(cap)s%(cluster)s",
                  {'service_name': service_name, 'host': host,
                   'cap': capabilities,
                   'cluster': cluster_msg})

        self._no_capabilities_backends.discard(backend)
        if just_init:
            self._update_backend_state_map(cinder_context.get_admin_context())

    def notify_service_capabilities(self, service_name, backend, capabilities,
                                    timestamp):
        """Notify the ceilometer with updated volume stats"""
        if service_name != 'volume':
            return

        updated = []
        capa_new = self.service_states.get(backend, {})
        timestamp = timestamp or timeutils.utcnow()

        # Compare the capabilities and timestamps to decide notifying
        if not capa_new:
            updated = self._get_updated_pools(capa_new, capabilities)
        else:
            if timestamp > self.service_states[backend]["timestamp"]:
                updated = self._get_updated_pools(
                    self.service_states[backend], capabilities)
                if not updated:
                    updated = self._get_updated_pools(
                        self.service_states_last_update.get(backend, {}),
                        self.service_states.get(backend, {}))

        if updated:
            capab_copy = dict(capabilities)
            capab_copy["timestamp"] = timestamp
            # If capabilities changes, notify and record the capabilities.
            self.service_states_last_update[backend] = capab_copy
            self.get_usage_and_notify(capabilities, updated, backend,
                                      timestamp)

    def has_all_capabilities(self):
        return len(self._no_capabilities_backends) == 0

    def _is_just_initialized(self):
        return not self.service_states_last_update

    def first_receive_capabilities(self):
        return (not self._is_just_initialized() and
                len(set(self.backend_state_map)) > 0 and
                len(self._no_capabilities_backends) == 0)

    def _update_backend_state_map(self, context):

        # Get resource usage across the available volume nodes:
        topic = constants.VOLUME_TOPIC
        volume_services = objects.ServiceList.get_all(context,
                                                      {'topic': topic,
                                                       'disabled': False,
                                                       'frozen': False})
        active_backends = set()
        active_hosts = set()
        no_capabilities_backends = set()
        for service in volume_services.objects:
            host = service.host
            if not service.is_up:
                LOG.warning("volume service is down. (host: %s)", host)
                continue

            backend_key = service.service_topic_queue
            # We only pay attention to the first up service of a cluster since
            # they all refer to the same capabilities entry in service_states
            if backend_key in active_backends:
                active_hosts.add(host)
                continue

            # Capabilities may come from the cluster or the host if the service
            # has just been converted to a cluster service.
            capabilities = (self.service_states.get(service.cluster_name, None)
                            or self.service_states.get(service.host, None))
            if capabilities is None:
                no_capabilities_backends.add(backend_key)
                continue

            # Since the service could have been added or remove from a cluster
            backend_state = self.backend_state_map.get(backend_key, None)
            if not backend_state:
                backend_state = self.backend_state_cls(
                    host,
                    service.cluster_name,
                    capabilities=capabilities,
                    service=dict(service))
                self.backend_state_map[backend_key] = backend_state

            # update capabilities and attributes in backend_state
            backend_state.update_from_volume_capability(capabilities,
                                                        service=dict(service))
            active_backends.add(backend_key)

        self._no_capabilities_backends = no_capabilities_backends

        # remove non-active keys from backend_state_map
        inactive_backend_keys = set(self.backend_state_map) - active_backends
        for backend_key in inactive_backend_keys:
            # NOTE(geguileo): We don't want to log the removal of a host from
            # the map when we are removing it because it has been added to a
            # cluster.
            if backend_key not in active_hosts:
                LOG.info("Removing non-active backend: %(backend)s from "
                         "scheduler cache.", {'backend': backend_key})
            del self.backend_state_map[backend_key]

    def revert_volume_consumed_capacity(self, pool_name, size):
        for backend_key, state in self.backend_state_map.items():
            for key in state.pools:
                pool_state = state.pools[key]
                if pool_name == '#'.join([backend_key, pool_state.pool_name]):
                    pool_state.consume_from_volume({'size': -size},
                                                   update_time=False)

    def get_all_backend_states(self, context):
        """Returns a dict of all the backends the HostManager knows about.

        Each of the consumable resources in BackendState are
        populated with capabilities scheduler received from RPC.

        For example:
          {'192.168.1.100': BackendState(), ...}
        """

        self._update_backend_state_map(context)

        # build a pool_state map and return that map instead of
        # backend_state_map
        all_pools = {}
        for backend_key, state in self.backend_state_map.items():
            for key in state.pools:
                pool = state.pools[key]
                # use backend_key.pool_name to make sure key is unique
                pool_key = '.'.join([backend_key, pool.pool_name])
                all_pools[pool_key] = pool

        return all_pools.values()

    def _filter_pools_by_volume_type(self, context, volume_type, pools):
        """Return the pools filtered by volume type specs"""

        # wrap filter properties only with volume_type
        filter_properties = {
            'context': context,
            'volume_type': volume_type,
            'resource_type': volume_type,
            'qos_specs': volume_type.get('qos_specs'),
        }

        filtered = self.get_filtered_backends(pools.values(),
                                              filter_properties)

        # filter the pools by value
        return {k: v for k, v in pools.items() if v in filtered}

    def get_pools(self, context, filters=None):
        """Returns a dict of all pools on all hosts HostManager knows about."""

        self._update_backend_state_map(context)

        all_pools = {}
        name = volume_type = None
        if filters:
            name = filters.pop('name', None)
            volume_type = filters.pop('volume_type', None)

        for backend_key, state in self.backend_state_map.items():
            for key in state.pools:
                filtered = False
                pool = state.pools[key]
                # use backend_key.pool_name to make sure key is unique
                pool_key = volume_utils.append_host(backend_key,
                                                    pool.pool_name)
                new_pool = dict(name=pool_key)
                new_pool.update(dict(capabilities=pool.capabilities))

                if name and new_pool.get('name') != name:
                    continue

                if filters:
                    # filter all other items in capabilities
                    for (attr, value) in filters.items():
                        cap = new_pool.get('capabilities').get(attr)
                        if not self._equal_after_convert(cap, value):
                            filtered = True
                            break

                if not filtered:
                    all_pools[pool_key] = pool

        # filter pools by volume type
        if volume_type:
            volume_type = volume_types.get_by_name_or_id(
                context, volume_type)
            all_pools = (
                self._filter_pools_by_volume_type(context,
                                                  volume_type,
                                                  all_pools))

        # encapsulate pools in format:{name: XXX, capabilities: XXX}
        return [dict(name=key, capabilities=value.capabilities)
                for key, value in all_pools.items()]

    def get_usage_and_notify(self, capa_new, updated_pools, host, timestamp):
        context = cinder_context.get_admin_context()
        usage = self._get_usage(capa_new, updated_pools, host, timestamp)

        self._notify_capacity_usage(context, usage)

    def _get_usage(self, capa_new, updated_pools, host, timestamp):
        pools = capa_new.get('pools')
        usage = []
        if pools and isinstance(pools, list):
            backend_usage = dict(type='backend',
                                 name_to_id=host,
                                 total=0,
                                 free=0,
                                 allocated=0,
                                 provisioned=0,
                                 virtual_free=0,
                                 reported_at=timestamp)

            # Process the usage.
            for pool in pools:
                pool_usage = self._get_pool_usage(pool, host, timestamp)
                if pool_usage:
                    backend_usage["total"] += pool_usage["total"]
                    backend_usage["free"] += pool_usage["free"]
                    backend_usage["allocated"] += pool_usage["allocated"]
                    backend_usage["provisioned"] += pool_usage["provisioned"]
                    backend_usage["virtual_free"] += pool_usage["virtual_free"]
                # Only the updated pool is reported.
                if pool in updated_pools:
                    usage.append(pool_usage)
            usage.append(backend_usage)
        return usage

    def _get_pool_usage(self, pool, host, timestamp):
        total = pool["total_capacity_gb"]
        free = pool["free_capacity_gb"]

        unknowns = ["unknown", "infinite", None]
        if (total in unknowns) or (free in unknowns):
            return {}

        allocated = pool["allocated_capacity_gb"]
        provisioned = pool["provisioned_capacity_gb"]
        reserved = pool["reserved_percentage"]
        ratio = utils.calculate_max_over_subscription_ratio(
            pool, CONF.max_over_subscription_ratio)
        support = pool["thin_provisioning_support"]

        virtual_free = utils.calculate_virtual_free_capacity(
            total,
            free,
            provisioned,
            support,
            ratio,
            reserved,
            support)

        pool_usage = dict(
            type='pool',
            name_to_id='#'.join([host, pool['pool_name']]),
            total=float(total),
            free=float(free),
            allocated=float(allocated),
            provisioned=float(provisioned),
            virtual_free=float(virtual_free),
            reported_at=timestamp)

        return pool_usage

    def _get_updated_pools(self, old_capa, new_capa):
        # Judge if the capabilities should be reported.

        new_pools = new_capa.get('pools', [])
        if not new_pools:
            return []

        if isinstance(new_pools, list):
            # If the volume_stats is not well prepared, don't notify.
            if not all(
                    self.REQUIRED_KEYS.issubset(pool) for pool in new_pools):
                return []
        else:
            LOG.debug("The reported capabilities are not well structured...")
            return []

        old_pools = old_capa.get('pools', [])
        if not old_pools:
            return new_pools

        updated_pools = []

        newpools = {}
        oldpools = {}
        for new_pool in new_pools:
            newpools[new_pool['pool_name']] = new_pool

        for old_pool in old_pools:
            oldpools[old_pool['pool_name']] = old_pool

        for key in newpools:
            if key in oldpools.keys():
                for k in self.REQUIRED_KEYS:
                    if newpools[key][k] != oldpools[key][k]:
                        updated_pools.append(newpools[key])
                        break
            else:
                updated_pools.append(newpools[key])

        return updated_pools

    def _notify_capacity_usage(self, context, usage):
        if usage:
            for u in usage:
                volume_utils.notify_about_capacity_usage(
                    context, u, u['type'], None, None)
        LOG.debug("Publish storage capacity: %s.", usage)

    def _equal_after_convert(self, capability, value):

        if isinstance(value, type(capability)) or capability is None:
            return value == capability

        if isinstance(capability, bool):
            return capability == strutils.bool_from_string(value)

        # We can not check or convert value parameter's type in
        # anywhere else.
        # If the capability and value are not in the same type,
        # we just convert them into string to compare them.
        return str(value) == str(capability)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import inspect
from itertools import chain

from oslo_config import cfg

from cinder import exception
from cinder.message import message_field
from cinder import test

CONF = cfg.CONF


@ddt.ddt
class MessageFieldTest(test.TestCase):
    def test_unique_action_ids(self):
        """Assert that no action_id is duplicated."""
        action_ids = [x[0] for x in message_field.Action.ALL]
        self.assertEqual(len(action_ids), len(set(action_ids)))

    def test_all_action_fields_in_ALL(self):
        """Assert that all and only defined fields are in the ALL tuple"""
        defined_fields = [k for k in message_field.Action.__dict__.keys()
                          if k != 'ALL' and not k.startswith('__')]
        for d in defined_fields:
            self.assertIn(getattr(message_field.Action, d),
                          message_field.Action.ALL)
        self.assertEqual(len(message_field.Action.ALL),
                         len(defined_fields))

    def test_unique_detail_ids(self):
        """Assert that no detail_id is duplicated."""
        detail_ids = [x[0] for x in message_field.Detail.ALL]
        self.assertEqual(len(detail_ids), len(set(detail_ids)))

    def test_all_detail_fields_in_ALL(self):
        """Assert that all and only defined fields are in the ALL tuple"""
        defined_fields = [k for k in message_field.Detail.__dict__.keys()
                          if k != 'ALL' and not k.startswith('__')
                          and k != 'EXCEPTION_DETAIL_MAPPINGS']
        for d in defined_fields:
            self.assertIn(getattr(message_field.Detail, d),
                          message_field.Detail.ALL)
        self.assertEqual(len(message_field.Detail.ALL),
                         len(defined_fields))

    known_exceptions = [
        name for name, _ in
        inspect.getmembers(exception, inspect.isclass)]
    mapped_exceptions = list(chain.from_iterable(
        message_field.Detail.EXCEPTION_DETAIL_MAPPINGS.values()))

    @ddt.idata(mapped_exceptions)
    def test_exception_detail_map_no_unknown_exceptions(self, exc):
        """Assert that only known exceptions are in the map."""
        self.assertIn(exc, self.known_exceptions)


@ddt.ddt
class MessageFieldFunctionsTest(test.TestCase):

    @ddt.data({'id': '001', 'content': 'schedule allocate volume'},
              {'id': '002', 'content': 'attach volume'},
              {'id': 'invalid', 'content': None})
    @ddt.unpack
    def test_translate_action(self, id, content):
        result = message_field.translate_action(id)
        if content is None:
            content = 'unknown action'
        self.assertEqual(content, result)

    @ddt.data({'id': '001',
               'content': 'An unknown error occurred.'},
              {'id': '002',
               'content': 'Driver is not initialized at present.'},
              {'id': 'invalid', 'content': None})
    @ddt.unpack
    def test_translate_detail(self, id, content):
        result = message_field.translate_detail(id)
        if content is None:
            content = 'An unknown error occurred.'
        self.assertEqual(content, result)

    @ddt.data({'exception': exception.DriverNotInitialized(),
               'detail': '',
               'expected': '002'},
              {'exception': exception.CinderException(),
               'detail': '',
               'expected': '001'},
              {'exception': exception.CinderException(),
               'detail': message_field.Detail.QUOTA_EXCEED,
               'expected': '007'},
              {'exception': '', 'detail': message_field.Detail.QUOTA_EXCEED,
               'expected': '007'})
    @ddt.unpack
    def translate_detail_id(self, exception, detail, expected):
        result = message_field.translate_detail_id(exception, detail)
        self.assertEqual(expected, result)



# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder.tests.unit.volume.drivers.dell_emc.vnx import fake_enum


class VNXSystem(object):
    pass


class VNXEnum(fake_enum.Enum):
    pass


class VNXSPEnum(VNXEnum):
    SP_A = 'SP A'
    SP_B = 'SP B'
    CONTROL_STATION = 'Celerra'


class VNXProvisionEnum(VNXEnum):
    # value of spec "provisioning:type"
    THIN = 'thin'
    THICK = 'thick'
    COMPRESSED = 'compressed'
    DEDUPED = 'deduplicated'


class VNXMigrationRate(VNXEnum):
    LOW = 'low'
    MEDIUM = 'medium'
    HIGH = 'high'
    ASAP = 'asap'


class VNXTieringEnum(VNXEnum):
    NONE = 'none'
    HIGH_AUTO = 'starthighthenauto'
    AUTO = 'auto'
    HIGH = 'highestavailable'
    LOW = 'lowestavailable'
    NO_MOVE = 'nomovement'


class VNXMirrorViewRecoveryPolicy(VNXEnum):
    MANUAL = 'manual'
    AUTO = 'automatic'


class VNXMirrorViewSyncRate(VNXEnum):
    HIGH = 'high'
    MEDIUM = 'medium'
    LOW = 'low'


class VNXMirrorImageState(VNXEnum):
    SYNCHRONIZED = 'Synchronized'
    OUT_OF_SYNC = 'Out-of-Sync'
    SYNCHRONIZING = 'Synchronizing'
    CONSISTENT = 'Consistent'
    SCRAMBLED = 'Scrambled'
    INCOMPLETE = 'Incomplete'
    LOCAL_ONLY = 'Local Only'
    EMPTY = 'Empty'


VNXCtrlMethod = fake_enum.VNXCtrlMethod



# Copyright 2011 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import iso8601
from oslo_config import cfg
from oslo_serialization import jsonutils
from six.moves import http_client
import webob

from cinder.api import extensions
from cinder.api.v3 import router
from cinder.tests.functional import functional_helpers


NS = "{http://docs.openstack.org/common/api/v1.0}"
CONF = cfg.CONF


class ExtensionTestCase(functional_helpers._FunctionalTestBase):
    def _get_flags(self):
        f = super(ExtensionTestCase, self)._get_flags()
        f['osapi_volume_extension'] = {'v': CONF.osapi_volume_extension[:]}
        f['osapi_volume_extension']['v'].append(
            'cinder.tests.functional.api.foxinsocks.Foxinsocks')
        return f


class ExtensionsTest(ExtensionTestCase):
    def test_get_foxnsocks(self):
        """Simple check that fox-n-socks works."""
        response = self.api.api_request('/foxnsocks')
        foxnsocks = response.text
        self.assertEqual('Try to say this Mr. Knox, sir...', foxnsocks)


class ExtensionControllerTest(ExtensionTestCase):

    def setUp(self):
        super(ExtensionControllerTest, self).setUp()
        self.ext_list = ["TypesManage", "TypesExtraSpecs", ]
        self.ext_list.sort()

    def test_list_extensions_json(self):
        app = router.APIRouter()
        request = webob.Request.blank("/fake/extensions")
        response = request.get_response(app)
        self.assertEqual(http_client.OK, response.status_int)

        # Make sure we have all the extensions, extra extensions being OK.
        data = jsonutils.loads(response.body)
        names = [str(x['name']) for x in data['extensions']
                 if str(x['name']) in self.ext_list]
        names.sort()
        self.assertEqual(self.ext_list, names)

        # Ensure all the timestamps are valid according to iso8601
        for ext in data['extensions']:
            iso8601.parse_date(ext['updated'])

        # Make sure that at least Fox in Sox is correct.
        (fox_ext, ) = [
            x for x in data['extensions'] if x['alias'] == 'FOXNSOX']
        self.assertEqual(
            {'name': 'Fox In Socks',
             'updated': '2011-01-22T13:25:27-06:00',
             'description': 'The Fox In Socks Extension.',
             'alias': 'FOXNSOX',
             'links': []},
            fox_ext)

        for ext in data['extensions']:
            url = '/fake/extensions/%s' % ext['alias']
            request = webob.Request.blank(url)
            response = request.get_response(app)
            output = jsonutils.loads(response.body)
            self.assertEqual(ext['alias'], output['extension']['alias'])

    def test_get_extension_json(self):
        app = router.APIRouter()
        request = webob.Request.blank("/fake/extensions/FOXNSOX")
        response = request.get_response(app)
        self.assertEqual(http_client.OK, response.status_int)

        data = jsonutils.loads(response.body)
        self.assertEqual(
            {"name": "Fox In Socks",
             "updated": "2011-01-22T13:25:27-06:00",
             "description": "The Fox In Socks Extension.",
             "alias": "FOXNSOX",
             "links": []}, data['extension'])

    def test_get_non_existing_extension_json(self):
        app = router.APIRouter()
        request = webob.Request.blank("/fake/extensions/4")
        response = request.get_response(app)
        self.assertEqual(http_client.NOT_FOUND, response.status_int)


class StubExtensionManager(object):
    """Provides access to Tweedle Beetles."""

    name = "Tweedle Beetle Extension"
    alias = "TWDLBETL"

    def __init__(self, resource_ext=None, action_ext=None, request_ext=None,
                 controller_ext=None):
        self.resource_ext = resource_ext
        self.controller_ext = controller_ext
        self.extra_resource_ext = None

    def get_resources(self):
        resource_exts = []
        if self.resource_ext:
            resource_exts.append(self.resource_ext)
        if self.extra_resource_ext:
            resource_exts.append(self.extra_resource_ext)
        return resource_exts

    def get_controller_extensions(self):
        controller_extensions = []
        if self.controller_ext:
            controller_extensions.append(self.controller_ext)
        return controller_extensions


class ExtensionControllerIdFormatTest(ExtensionTestCase):

    def _bounce_id(self, test_id):

        class BounceController(object):
            def show(self, req, id):
                return id
        res_ext = extensions.ResourceExtension('bounce',
                                               BounceController())
        manager = StubExtensionManager(res_ext)
        app = router.APIRouter(manager)
        request = webob.Request.blank("/fake/bounce/%s" % test_id)
        response = request.get_response(app)
        return response.body

    def test_id_with_json_format(self):
        result = self._bounce_id('foo.json')
        self.assertEqual(b'foo', result)

    def test_id_with_bad_format(self):
        result = self._bounce_id('foo.bad')
        self.assertEqual(b'foo.bad', result)



# Copyright (c) 2017 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from cinder.policies import base


CREATE_POLICY = "volume_extension:types_extra_specs:create"
DELETE_POLICY = "volume_extension:types_extra_specs:delete"
GET_ALL_POLICY = "volume_extension:types_extra_specs:index"
GET_POLICY = "volume_extension:types_extra_specs:show"
UPDATE_POLICY = "volume_extension:types_extra_specs:update"


type_extra_specs_policies = [
    policy.DocumentedRuleDefault(
        name=GET_ALL_POLICY,
        check_str=base.RULE_ADMIN_API,
        description="List type extra specs.",
        operations=[
            {
                'method': 'GET',
                'path': '/types/{type_id}/extra_specs'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=CREATE_POLICY,
        check_str=base.RULE_ADMIN_API,
        description="Create type extra specs.",
        operations=[
            {
                'method': 'POST',
                'path': '/types/{type_id}/extra_specs'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=GET_POLICY,
        check_str=base.RULE_ADMIN_API,
        description="Show one specified type extra specs.",
        operations=[
            {
                'method': 'GET',
                'path': '/types/{type_id}/extra_specs/{extra_spec_key}'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=UPDATE_POLICY,
        check_str=base.RULE_ADMIN_API,
        description="Update type extra specs.",
        operations=[
            {
                'method': 'PUT',
                'path': '/types/{type_id}/extra_specs/{extra_spec_key}'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=DELETE_POLICY,
        check_str=base.RULE_ADMIN_API,
        description="Delete type extra specs.",
        operations=[
            {
                'method': 'DELETE',
                'path': '/types/{type_id}/extra_specs/{extra_spec_key}'
            }
        ]),
]


def list_rules():
    return type_extra_specs_policies
